@@ -21,16 +21,18 @@
 #ifndef OPM_AQUIFERCT_HEADER_INCLUDED
 #define OPM_AQUIFERCT_HEADER_INCLUDED
 
-#include <opm/simulators/aquifers/AquiferAnalytical.hpp>
+#include <opm/common/utility/numeric/linearInterpolation.hpp>
 
 #include <opm/input/eclipse/EclipseState/Aquifer/AquiferCT.hpp>
 
 #include <opm/output/data/Aquifer.hpp>
 
-#include <exception>
-#include <memory>
+#include <opm/simulators/aquifers/AquiferAnalytical.hpp>
+
+#include <cstddef>
 #include <stdexcept>
 #include <utility>
+#include <vector>
 
 namespace Opm
 {
@@ -53,15 +55,15 @@ public:
     using typename Base::ElementMapper;
 
     AquiferCarterTracy(const std::vector<Aquancon::AquancCell>& connections,
-                       const Simulator& ebosSimulator,
+                       const Simulator& simulator,
                        const AquiferCT::AQUCT_data& aquct_data)
-        : Base(aquct_data.aquiferID, connections, ebosSimulator)
+        : Base(aquct_data.aquiferID, connections, simulator)
         , aquct_data_(aquct_data)
     {}
 
-    static AquiferCarterTracy serializationTestObject(const Simulator& ebosSimulator)
+    static AquiferCarterTracy serializationTestObject(const Simulator& simulator)
     {
-        AquiferCarterTracy result({}, ebosSimulator, {});
+        AquiferCarterTracy result({}, simulator, {});
 
         result.pressure_previous_ = {1.0, 2.0, 3.0};
         result.pressure_current_ = {4.0, 5.0};
@@ -78,10 +80,10 @@ public:
     void endTimeStep() override
     {
         for (const auto& q : this->Qai_) {
-            this->W_flux_ += q * this->ebos_simulator_.timeStepSize();
+            this->W_flux_ += q * this->simulator_.timeStepSize();
         }
         this->fluxValue_ = this->W_flux_.value();
-        const auto& comm = this->ebos_simulator_.vanguard().grid().comm();
+        const auto& comm = this->simulator_.vanguard().grid().comm();
         comm.sum(&this->fluxValue_, 1);
     }
 
@@ -104,7 +106,7 @@ public:
         aquCT->dimensionless_pressure = this->dimensionless_pressure_;
         aquCT->influxConstant = this->aquct_data_.influxConstant();
 
-        if (!this->co2store_()) {
+        if (!this->co2store_or_h2store_()) {
             aquCT->timeConstant = this->aquct_data_.timeConstant();
             aquCT->waterDensity = this->aquct_data_.waterDensity();
             aquCT->waterViscosity = this->aquct_data_.waterViscosity();
@@ -207,7 +209,7 @@ protected:
     }
 
     // This function implements Eq 5.7 of the EclipseTechnicalDescription
-    inline void calculateInflowRate(int idx, const Simulator& simulator) override
+    void calculateInflowRate(int idx, const Simulator& simulator) override
     {
         const auto [a, b] = this->calculateEqnConstants(idx, simulator);
 
@@ -215,34 +217,16 @@ protected:
             (a - b*(this->pressure_current_.at(idx) - this->pressure_previous_.at(idx)));
     }
 
-    inline void calculateAquiferConstants() override
+    void calculateAquiferConstants() override
     {
-        if(this->co2store_()) {
-             const auto press = this->aquct_data_.initial_pressure.value();
-             Scalar temp = FluidSystem::reservoirTemperature();
-             if (this->aquct_data_.initial_temperature.has_value())
-                 temp = this->aquct_data_.initial_temperature.value();
+        this->Tc_ = this->co2store_or_h2store_()
+            ? this->timeConstantCO2Store()
+            : this->aquct_data_.timeConstant();
 
-             Scalar waterViscosity = 0.;
-             if (FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx)) {
-                 Scalar rs = 0.0; // no dissolved CO2
-                 waterViscosity = FluidSystem::oilPvt().viscosity(pvtRegionIdx(), temp, press, rs);
-             } else if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
-                 Scalar salt = 0.;
-                 Scalar rsw = 0.;
-                 waterViscosity = FluidSystem::waterPvt().viscosity(pvtRegionIdx(), temp, press, rsw, salt);
-             } else {
-                 OPM_THROW(std::runtime_error, "water or oil phase is needed to run CO2Store.");
-             }
-             const auto x = this->aquct_data_.porosity * this->aquct_data_.total_compr * this->aquct_data_.inner_radius * this->aquct_data_.inner_radius;
-             this->Tc_  = waterViscosity * x / this->aquct_data_.permeability;
-        } else {
-             this->Tc_ = this->aquct_data_.timeConstant();
-        }
         this->beta_ = this->aquct_data_.influxConstant();
     }
 
-    inline void calculateAquiferCondition() override
+    void calculateAquiferCondition() override
     {
         if (this->solution_set_from_restart_) {
             return;
@@ -252,46 +236,90 @@ protected:
             this->aquct_data_.initial_pressure =
                 this->calculateReservoirEquilibrium();
 
-            const auto& tables = this->ebos_simulator_.vanguard()
+            const auto& tables = this->simulator_.vanguard()
                 .eclState().getTableManager();
 
             this->aquct_data_.finishInitialisation(tables);
         }
 
         this->pa0_ = this->aquct_data_.initial_pressure.value();
-        if (this->aquct_data_.initial_temperature.has_value())
+        if (this->aquct_data_.initial_temperature.has_value()) {
             this->Ta0_ = this->aquct_data_.initial_temperature.value();
-
-        if(this->co2store_()) {
-            const auto press = this->aquct_data_.initial_pressure.value();
-
-            Scalar temp = FluidSystem::reservoirTemperature();
-            if (this->aquct_data_.initial_temperature.has_value())
-                temp = this->aquct_data_.initial_temperature.value();
-
-            Scalar waterDensity = 0.;
-            if (FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx)) {
-                Scalar rs = 0.0; // no dissolved CO2
-                waterDensity = FluidSystem::oilPvt().inverseFormationVolumeFactor(pvtRegionIdx(), temp, press, rs)
-                               * FluidSystem::oilPvt().oilReferenceDensity(pvtRegionIdx());
-            } else if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
-                Scalar salinity = 0.;
-                Scalar rsw = 0.;
-                waterDensity = FluidSystem::waterPvt().inverseFormationVolumeFactor(pvtRegionIdx(), temp, press, rsw, salinity)
-                               * FluidSystem::waterPvt().waterReferenceDensity(pvtRegionIdx());
-            } else {
-                OPM_THROW(std::runtime_error, "water or oil phase is needed to run CO2Store.");
-            }
-            this->rhow_  = waterDensity;
-        } else {
-            this->rhow_ = this->aquct_data_.waterDensity();
         }
+
+        this->rhow_ = this->co2store_or_h2store_()
+            ? this->waterDensityCO2Store()
+            : this->aquct_data_.waterDensity();
     }
 
-    virtual Scalar aquiferDepth() const override
+    Scalar aquiferDepth() const override
     {
         return this->aquct_data_.datum_depth;
     }
+
+private:
+    Scalar timeConstantCO2Store() const
+    {
+        const auto press = this->aquct_data_.initial_pressure.value();
+        const auto temp = this->reservoirTemperatureCO2Store();
+
+        auto waterViscosity = Scalar { 0 };
+        if (FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx)) {
+            const auto rs = Scalar { 0 }; // no dissolved CO2
+            waterViscosity = FluidSystem::oilPvt()
+                .viscosity(pvtRegionIdx(), temp, press, rs);
+        }
+        else if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
+            const auto salt = Scalar { 0 };
+            const auto rsw = Scalar { 0 };
+            waterViscosity = FluidSystem::waterPvt()
+                .viscosity(pvtRegionIdx(), temp, press, rsw, salt);
+        }
+        else {
+            OPM_THROW(std::runtime_error, "water or oil phase is needed to run CO2Store.");
+        }
+
+        const auto x = this->aquct_data_.porosity * this->aquct_data_.total_compr
+            * this->aquct_data_.inner_radius * this->aquct_data_.inner_radius;
+
+        return waterViscosity * x / this->aquct_data_.permeability;
+    }
+
+    Scalar waterDensityCO2Store() const
+    {
+        const auto press = this->aquct_data_.initial_pressure.value();
+        const auto temp = this->reservoirTemperatureCO2Store();
+
+        if (FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx)) {
+            const auto& pvt = FluidSystem::oilPvt();
+            const auto reg = this->pvtRegionIdx();
+
+            const auto rs = Scalar { 0 }; // no dissolved CO2
+            return pvt.inverseFormationVolumeFactor(reg, temp, press, rs)
+                * pvt.oilReferenceDensity(reg);
+        }
+        else if (FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx)) {
+            const auto& pvt = FluidSystem::waterPvt();
+            const auto reg = this->pvtRegionIdx();
+
+            const auto salinity = Scalar { 0 };
+            const auto rsw = Scalar { 0 };
+
+            return pvt.inverseFormationVolumeFactor(reg, temp, press, rsw, salinity)
+                * pvt.waterReferenceDensity(reg);
+        }
+        else {
+            OPM_THROW(std::runtime_error, "water or oil phase is needed to run CO2Store.");
+        }
+    }
+
+    Scalar reservoirTemperatureCO2Store() const
+    {
+        return this->aquct_data_.initial_temperature.has_value()
+            ? this->aquct_data_.initial_temperature.value()
+            : FluidSystem::reservoirTemperature();
+    }
+
 }; // class AquiferCarterTracy
 
 } // namespace Opm
