@@ -29,6 +29,7 @@
 #include <opm/simulators/utils/readDeck.hpp>
 
 #include <opm/common/ErrorMacros.hpp>
+#include <opm/common/TimingMacros.hpp>
 #include <opm/common/OpmLog/EclipsePRTLog.hpp>
 #include <opm/common/OpmLog/OpmLog.hpp>
 #include <opm/common/utility/OpmInputError.hpp>
@@ -75,6 +76,7 @@
 #include <filesystem>
 #include <functional>
 #include <memory>
+#include <regex>
 #include <sstream>
 #include <stdexcept>
 #include <unordered_map>
@@ -153,6 +155,11 @@ namespace {
                  std::move(python), outputInterval, init_state);
         }
 
+        // Read network pressures from restart
+        if (rst_state.network.isActive()) {
+            eclipseState.loadRestartNetworkPressures(rst_state.network);
+        }
+
         udqState = std::make_unique<Opm::UDQState>
             ((*schedule)[0].udq().params().undefinedValue());
         udqState->load_rst(rst_state);
@@ -240,6 +247,7 @@ namespace {
                       const std::optional<int>&            outputInterval,
                       Opm::ErrorGuard&                     errorGuard)
     {
+        OPM_TIMEBLOCK(readDeck);
         if (((schedule == nullptr) || (summaryConfig == nullptr)) &&
             (parseContext == nullptr))
         {
@@ -253,6 +261,7 @@ namespace {
                                        *parseContext, treatCriticalAsNonCritical, errorGuard);
 
         if (eclipseState == nullptr) {
+            OPM_TIMEBLOCK(createEclState);
             eclipseState = createEclipseState(comm, deck);
         }
 
@@ -323,27 +332,39 @@ namespace {
     }
 #endif
 
-    bool gridHasValidCellGeometry(const Opm::EclipseGrid& inputGrid,
-                                  const Opm::UnitSystem&  usys)
+    std::pair<bool, std::string>
+    gridHasValidCellGeometry(const Opm::EclipseGrid& inputGrid,
+                             const Opm::UnitSystem&  usys)
     {
         const auto numActive = inputGrid.getNumActive();
 
+        if (numActive == 0)
+        {
+            return {false, R"(Input grid has no active cells at all.
+Please check geometry keywords and modifications to e.g. pore volume,
+especially if grid is imported through GDFILE.)"};
+        }
+
         for (auto activeCell = 0*numActive; activeCell < numActive; ++activeCell) {
             if (inputGrid.isValidCellGeomtry(inputGrid.getGlobalIndex(activeCell), usys)) {
-                return true;
+                return {true, ""};
             }
         }
 
-        return false;
+        return {false, R"(No active cell in input grid has valid/finite cell geometry
+Please check geometry keywords and modifications to e.g. pore volume,
+especially if the grid is imported through GDFILE.)"};
     }
 
-    bool gridHasValidCellGeometry(Opm::Parallel::Communication comm,
+    std::pair<bool, std::string>
+    gridHasValidCellGeometry(Opm::Parallel::Communication comm,
                                   const Opm::EclipseState&     eclipseState)
     {
         bool hasValidCells = false;
+        std::string message;
 
         if (comm.rank() == 0) {
-            hasValidCells =
+            std::tie(hasValidCells, message) =
                 gridHasValidCellGeometry(eclipseState.getInputGrid(),
                                          eclipseState.getDeckUnitSystem());
         }
@@ -358,7 +379,7 @@ namespace {
         }
 #endif  // HAVE_MPI
 
-        return hasValidCells;
+        return {hasValidCells, message};
     }
 }
 
@@ -382,15 +403,28 @@ void Opm::ensureOutputDirExists(const std::string& cmdline_output_dir)
     }
 }
 
+void Opm::prepareResultOutputDirectory(const std::string&           baseName,
+                                       const std::filesystem::path& outputDir)  
+{   
+    //Loop over all files in the output directory and subdirectories and delete them if their name is baseName + a correct extension
+    std::regex r(baseName + R"(\.(F?(DBG|E?GRID|INIT|PRT|RFT|SMSPEC|UNSMRY|UNRST)|([ABCFGHSTUXYZ]\d{4})|(INFOSTEP|INFOITER|OPMRST)))");
+    for (auto& file : std::filesystem::recursive_directory_iterator(outputDir)) {
+        std::string fileName = file.path().filename();
+        if (std::regex_match(fileName, r)) {
+            std::filesystem::remove(file);
+        }
+    }
+}
+
 // Setup the OpmLog backends
 Opm::FileOutputMode
-Opm::setupLogging(const int          mpi_rank_,
-                  const std::string& deck_filename,
-                  const std::string& cmdline_output_dir,
-                  const std::string& cmdline_output,
-                  const bool         output_cout_,
-                  const std::string& stdout_log_id,
-                  const bool         allRanksDbgLog)
+Opm::setupLogging(Parallel::Communication& comm,
+                  const std::string&       deck_filename,
+                  const std::string&       cmdline_output_dir,
+                  const std::string&       cmdline_output,
+                  const bool               output_cout_,
+                  const std::string&       stdout_log_id,
+                  const bool               allRanksDbgLog)
 {
     if (!cmdline_output_dir.empty()) {
         ensureOutputDirExists(cmdline_output_dir);
@@ -419,15 +453,24 @@ Opm::setupLogging(const int          mpi_rank_,
             : std::filesystem::current_path().generic_string();
     }
 
+    // Cleans up the result output directory, we only do this on process 0
+    if (comm.rank() == 0) { 
+        prepareResultOutputDirectory(baseName, output_dir);
+    }
+    //... and the other processes need to wait for this to be finished.
+#if HAVE_MPI
+    MPI_Barrier(comm);
+#endif
+
     logFileStream << output_dir << "/" << baseName;
     debugFileStream << output_dir << "/" << baseName;
 
-    if (mpi_rank_ != 0) {
+    if (comm.rank() != 0) {
         // Added rank to log file for non-zero ranks.
         // This prevents message loss.
-        debugFileStream << "." << mpi_rank_;
+        debugFileStream << "." << comm.rank();
         // If the following file appears then there is a bug.
-        logFileStream << "." << mpi_rank_;
+        logFileStream << "." << comm.rank();
     }
     logFileStream << ".PRT";
     debugFileStream << ".DBG";
@@ -449,7 +492,7 @@ Opm::setupLogging(const int          mpi_rank_,
             std::cerr << "Value " << cmdline_output
                       << " is not a recognized output mode. Using \"all\" instead.\n";
         }
-        if (!allRanksDbgLog && mpi_rank_ != 0)
+        if (!allRanksDbgLog && comm.rank() != 0)
         {
             output = FileOutputMode::OUTPUT_NONE;
         }
@@ -468,7 +511,7 @@ Opm::setupLogging(const int          mpi_rank_,
         Opm::OpmLog::addBackend("DEBUGLOG", debugLog);
     }
 
-    if (mpi_rank_ == 0) {
+    if (comm.rank() == 0) {
         std::shared_ptr<Opm::StreamLog> streamLog = std::make_shared<Opm::StreamLog>(std::cout, Opm::Log::StdoutMessageTypes);
         Opm::OpmLog::addBackend(stdout_log_id, streamLog);
         // Set a tag limit of 10 (no category limit). Will later in
@@ -512,10 +555,17 @@ void Opm::readDeck(Opm::Parallel::Communication    comm,
         const bool treatCriticalAsNonCritical = (parsingStrictness == "low");
         try {
             auto parseContext = setupParseContext(exitOnAllErrors);
+            if (treatCriticalAsNonCritical) { // Continue with invalid names if parsing strictness is set to low
+                parseContext->update(ParseContext::SCHEDULE_INVALID_NAME, InputErrorAction::WARN);
+            }
             readOnIORank(comm, deckFilename, parseContext.get(),
                          eclipseState, schedule, udqState, actionState, wtestState,
                          summaryConfig, std::move(python), initFromRestart,
                          checkDeck, treatCriticalAsNonCritical, outputInterval, *errorGuard);
+
+            // Update schedule so that re-parsing after actions use same strictness
+            assert(schedule);
+            schedule->treat_critical_as_non_critical(treatCriticalAsNonCritical);
         }
         catch (const OpmInputError& input_error) {
             failureMessage = input_error.what();
@@ -540,6 +590,7 @@ void Opm::readDeck(Opm::Parallel::Communication    comm,
     parseSuccess = comm.min(parseSuccess);
     try {
         if (parseSuccess) {
+            OPM_TIMEBLOCK(eclBcast);
             eclStateBroadcast(comm, *eclipseState, *schedule,
                               *summaryConfig, *udqState, *actionState, *wtestState);
         }
@@ -576,14 +627,12 @@ void Opm::readDeck(Opm::Parallel::Communication    comm,
 void Opm::verifyValidCellGeometry(Parallel::Communication comm,
                                   const EclipseState&     eclipseState)
 {
-    if (gridHasValidCellGeometry(comm, eclipseState)) {
+    auto [hasValidCells, message] = gridHasValidCellGeometry(comm, eclipseState);
+    if (hasValidCells) {
         return;
     }
 
-    throw std::invalid_argument {
-        R"(No active cell in input grid has valid/finite cell geometry
-Please check geometry keywords, especially if grid is imported through GDFILE)"
-    };
+    throw std::invalid_argument { message };
 }
 
 std::unique_ptr<Opm::ParseContext> Opm::setupParseContext(const bool strictParsing)
