@@ -24,48 +24,52 @@
 #ifndef OPM_BLACKOILWELLMODEL_HEADER_INCLUDED
 #define OPM_BLACKOILWELLMODEL_HEADER_INCLUDED
 
-#include <ebos/eclproblem.hh>
 #include <opm/common/OpmLog/OpmLog.hpp>
 
-#include <cassert>
 #include <cstddef>
 #include <map>
 #include <memory>
 #include <optional>
-#include <set>
 #include <string>
-#include <tuple>
-#include <unordered_map>
 #include <vector>
 
+#include <opm/input/eclipse/Schedule/Group/Group.hpp>
+#include <opm/input/eclipse/Schedule/Group/GuideRate.hpp>
 #include <opm/input/eclipse/Schedule/Schedule.hpp>
 #include <opm/input/eclipse/Schedule/Well/WellTestState.hpp>
-#include <opm/input/eclipse/Schedule/Group/GuideRate.hpp>
-#include <opm/input/eclipse/Schedule/Group/Group.hpp>
 
-#include <opm/simulators/timestepping/SimulatorReport.hpp>
+#include <opm/models/discretization/common/baseauxiliarymodule.hh>
+
 #include <opm/simulators/flow/countGlobalCells.hpp>
+#include <opm/simulators/flow/FlowBaseVanguard.hpp>
+#include <opm/simulators/flow/SubDomain.hpp>
+
+#include <opm/simulators/linalg/matrixblock.hh>
+
 #include <opm/simulators/wells/BlackoilWellModelGeneric.hpp>
 #include <opm/simulators/wells/BlackoilWellModelGuideRates.hpp>
-#include <opm/simulators/wells/GasLiftSingleWell.hpp>
-#include <opm/simulators/wells/GasLiftWellState.hpp>
-#include <opm/simulators/wells/GasLiftSingleWellGeneric.hpp>
-#include <opm/simulators/wells/GasLiftStage2.hpp>
 #include <opm/simulators/wells/GasLiftGroupInfo.hpp>
+#include <opm/simulators/wells/GasLiftSingleWell.hpp>
+#include <opm/simulators/wells/GasLiftSingleWellGeneric.hpp>
+#include <opm/simulators/wells/GasLiftWellState.hpp>
+#include <opm/simulators/wells/MultisegmentWell.hpp>
+#include <opm/simulators/wells/ParallelWBPCalculation.hpp>
+#include <opm/simulators/wells/ParallelWellInfo.hpp>
 #include <opm/simulators/wells/PerforationData.hpp>
-#include <opm/simulators/wells/VFPInjProperties.hpp>
-#include <opm/simulators/wells/VFPProdProperties.hpp>
-#include <opm/simulators/wells/WellState.hpp>
-#include <opm/simulators/wells/WGState.hpp>
 #include <opm/simulators/wells/RateConverter.hpp>
 #include <opm/simulators/wells/RegionAverageCalculator.hpp>
-#include <opm/simulators/wells/WellInterface.hpp>
 #include <opm/simulators/wells/StandardWell.hpp>
-#include <opm/simulators/wells/MultisegmentWell.hpp>
+#include <opm/simulators/wells/VFPInjProperties.hpp>
+#include <opm/simulators/wells/VFPProdProperties.hpp>
+#include <opm/simulators/wells/WGState.hpp>
 #include <opm/simulators/wells/WellGroupHelpers.hpp>
+#include <opm/simulators/wells/WellInterface.hpp>
 #include <opm/simulators/wells/WellProdIndexCalculator.hpp>
-#include <opm/simulators/wells/ParallelWellInfo.hpp>
+#include <opm/simulators/wells/WellState.hpp>
+
+#include <opm/simulators/timestepping/SimulatorReport.hpp>
 #include <opm/simulators/timestepping/gatherConvergenceReport.hpp>
+
 #include <dune/common/fmatrix.hh>
 #include <dune/istl/bcrsmatrix.hh>
 #include <dune/istl/matrixmatrix.hh>
@@ -92,7 +96,7 @@ namespace Opm {
         {
         public:
             // ---------      Types      ---------
-            typedef BlackoilModelParametersEbos<TypeTag> ModelParameters;
+            using ModelParameters = BlackoilModelParameters<TypeTag>;
 
             using Grid = GetPropType<TypeTag, Properties::Grid>;
             using EquilGrid = GetPropType<TypeTag, Properties::EquilGrid>;
@@ -137,7 +141,9 @@ namespace Opm {
             using AverageRegionalPressureType = RegionAverageCalculator::
                 AverageRegionalPressure<FluidSystem, std::vector<int> >;
 
-            BlackoilWellModel(Simulator& ebosSimulator);
+            using Domain = SubDomain<Grid>;
+
+            BlackoilWellModel(Simulator& simulator);
 
             void init();
             void initWellContainer(const int reportStepIdx) override;
@@ -155,12 +161,18 @@ namespace Opm {
             {}
 
             void linearize(SparseMatrixAdapter& jacobian, GlobalEqVector& res) override;
+            void linearizeDomain(const Domain& domain, SparseMatrixAdapter& jacobian, GlobalEqVector& res);
 
             void postSolve(GlobalEqVector& deltaX) override
             {
                 recoverWellSolutionAndUpdateWellState(deltaX);
             }
 
+            void postSolveDomain(GlobalEqVector& deltaX, const Domain& domain)
+            {
+                recoverWellSolutionAndUpdateWellStateDomain(deltaX, domain);
+            }
+
             /////////////
             // </ eWoms auxiliary module stuff>
             /////////////
@@ -184,7 +196,7 @@ namespace Opm {
             void beginEpisode()
             {
                 OPM_TIMEBLOCK(beginEpsiode);
-                beginReportStep(ebosSimulator_.episodeIndex());
+                beginReportStep(simulator_.episodeIndex());
             }
 
             void beginTimeStep();
@@ -192,8 +204,8 @@ namespace Opm {
             void beginIteration()
             {
                 OPM_TIMEBLOCK(beginIteration);
-                assemble(ebosSimulator_.model().newtonMethod().numIterations(),
-                         ebosSimulator_.timeStepSize());
+                assemble(simulator_.model().newtonMethod().numIterations(),
+                         simulator_.timeStepSize());
             }
 
             void endIteration()
@@ -202,7 +214,7 @@ namespace Opm {
             void endTimeStep()
             {
                 OPM_TIMEBLOCK(endTimeStep);
-                timeStepSucceeded(ebosSimulator_.time(), ebosSimulator_.timeStepSize());
+                timeStepSucceeded(simulator_.time(), simulator_.timeStepSize());
             }
 
             void endEpisode()
@@ -226,7 +238,7 @@ namespace Opm {
             void initFromRestartFile(const RestartValue& restartValues)
             {
                 initFromRestartFile(restartValues,
-                                    this->ebosSimulator_.vanguard().transferWTestState(),
+                                    this->simulator_.vanguard().transferWTestState(),
                                     grid().size(0),
                                     param_.use_multisegment_well_);
             }
@@ -241,13 +253,15 @@ namespace Opm {
             data::Wells wellData() const
             {
                 auto wsrpt = this->wellState()
-                    .report(ebosSimulator_.vanguard().globalCell().data(),
+                    .report(simulator_.vanguard().globalCell().data(),
                             [this](const int well_index) -> bool
                 {
                     return this->wasDynamicallyShutThisTimeStep(well_index);
                 });
 
-                this->assignWellTracerRates(wsrpt);
+                const auto& tracerRates = simulator_.problem().tracerModel().getWellTracerRates();
+                this->assignWellTracerRates(wsrpt, tracerRates);
+
 
                 BlackoilWellModelGuideRates(*this).assignWellGuideRates(wsrpt, this->reportStepIndex());
                 this->assignShutConnections(wsrpt, this->reportStepIndex());
@@ -255,6 +269,11 @@ namespace Opm {
                 return wsrpt;
             }
 
+            data::WellBlockAveragePressures wellBlockAveragePressures() const
+            {
+                return this->computeWellBlockAveragePressures();
+            }
+
             // subtract Binv(D)rw from r;
             void apply( BVector& r) const;
 
@@ -270,6 +289,11 @@ namespace Opm {
             // Check if well equations is converged.
             ConvergenceReport getWellConvergence(const std::vector<Scalar>& B_avg, const bool checkWellGroupControls = false) const;
 
+            // Check if well equations are converged locally.
+            ConvergenceReport getDomainWellConvergence(const Domain& domain,
+                                                       const std::vector<Scalar>& B_avg,
+                                                       DeferredLogger& local_deferredLogger) const;
+
             const SimulatorReportSingle& lastReport() const;
 
             void addWellContributions(SparseMatrixAdapter& jacobian) const;
@@ -281,7 +305,6 @@ namespace Opm {
             // called at the beginning of a report step
             void beginReportStep(const int time_step);
 
-            void updatePerforationIntensiveQuantities();
             // it should be able to go to prepareTimeStep(), however, the updateWellControls() and initPrimaryVariablesEvaluation()
             // makes it a little more difficult. unless we introduce if (iterationIdx != 0) to avoid doing the above functions
             // twice at the beginning of the time step
@@ -292,7 +315,10 @@ namespace Opm {
             // at the beginning of each time step (Not report step)
             void prepareTimeStep(DeferredLogger& deferred_logger);
             void initPrimaryVariablesEvaluation() const;
-            std::tuple<bool, bool, double> updateWellControls(DeferredLogger& deferred_logger);
+            void initPrimaryVariablesEvaluationDomain(const Domain& domain) const;
+
+            std::pair<bool, bool>
+            updateWellControls(const bool mandatory_network_balance, DeferredLogger& deferred_logger, const bool relax_network_tolerance = false);
 
             void updateAndCommunicate(const int reportStepIdx,
                                       const int iterationIdx,
@@ -308,12 +334,8 @@ namespace Opm {
 
             using PressureMatrix = Dune::BCRSMatrix<Opm::MatrixBlock<double, 1, 1>>;
 
-            int numLocalWellsEnd() const;
-
             void addWellPressureEquations(PressureMatrix& jacobian, const BVector& weights,const bool use_well_weights) const;
 
-            std::vector<std::vector<int>> getMaxWellConnections() const;
-
             void addWellPressureEquationsStruct(PressureMatrix& jacobian) const;
 
             void initGliftEclWellMap(GLiftEclWells &ecl_well_map);
@@ -324,25 +346,35 @@ namespace Opm {
                 return well_container_;
             }
 
-            int numLocalNonshutWells() const;
+            // prototype for assemble function for ASPIN solveLocal()
+            // will try to merge back to assemble() when done prototyping
+            void assembleDomain(const int iterationIdx,
+                                const double dt,
+                                const Domain& domain);
+            void updateWellControlsDomain(DeferredLogger& deferred_logger, const Domain& domain);
+
+            void logPrimaryVars() const;
+            std::vector<double> getPrimaryVarsDomain(const Domain& domain) const;
+            void setPrimaryVarsDomain(const Domain& domain, const std::vector<double>& vars);
+
+            void setupDomains(const std::vector<Domain>& domains);
 
         protected:
-            Simulator& ebosSimulator_;
+            Simulator& simulator_;
 
             // a vector of all the wells.
             std::vector<WellInterfacePtr> well_container_{};
 
             std::vector<bool> is_cell_perforated_{};
 
-            void initializeWellState(const int           timeStepIdx,
-                                     const SummaryState& summaryState);
+            void initializeWellState(const int timeStepIdx);
 
             // create the well container
-            void createWellContainer(const int time_step) override;
+            void createWellContainer(const int report_step) override;
 
             WellInterfacePtr
             createWellPointer(const int wellID,
-                              const int time_step) const;
+                              const int report_step) const;
 
             template <typename WellType>
             std::unique_ptr<WellType>
@@ -363,34 +395,72 @@ namespace Opm {
             std::unique_ptr<RateConverterType> rateConverter_{};
             std::map<std::string, std::unique_ptr<AverageRegionalPressureType>> regionalAveragePressureCalculator_{};
 
+            struct WBPCalcID
+            {
+                std::optional<typename std::vector<WellInterfacePtr>::size_type> openWellIdx_{};
+                std::size_t wbpCalcIdx_{};
+            };
+
+            std::vector<WBPCalcID> wbpCalcMap_{};
 
             SimulatorReportSingle last_report_{};
 
+            // Pre-step network solve at static reservoir conditions (group and well states might be updated)
+            void doPreStepNetworkRebalance(DeferredLogger& deferred_logger);
+
             // used to better efficiency of calcuation
             mutable BVector scaleAddRes_{};
 
             std::vector<Scalar> B_avg_{};
 
+            // Keep track of the domain of each well, if using subdomains.
+            std::map<std::string, int> well_domain_;
+
             const Grid& grid() const
-            { return ebosSimulator_.vanguard().grid(); }
+            { return simulator_.vanguard().grid(); }
 
             const EquilGrid& equilGrid() const
-            { return ebosSimulator_.vanguard().equilGrid(); }
+            { return simulator_.vanguard().equilGrid(); }
 
             const EclipseState& eclState() const
-            { return ebosSimulator_.vanguard().eclState(); }
+            { return simulator_.vanguard().eclState(); }
 
             // compute the well fluxes and assemble them in to the reservoir equations as source terms
             // and in the well equations.
             void assemble(const int iterationIdx,
                           const double dt);
-            bool assembleImpl(const int iterationIdx,
-                              const double dt,
-                              const std::size_t recursion_level,
-                              DeferredLogger& local_deferredLogger);
+
+            // well controls and network pressures affect each other and are solved in an iterative manner.
+            // the function handles one iteration of updating well controls and network pressures.
+            // it is possible to decouple the update of well controls and network pressures further.
+            // the returned two booleans are {continue_due_to_network, well_group_control_changed}, respectively
+            std::pair<bool, bool> updateWellControlsAndNetworkIteration(const bool mandatory_network_balance,
+                                                                        const bool relax_network_tolerance,
+                                                                        const double dt,
+                                                                        DeferredLogger& local_deferredLogger);
+
+            bool updateWellControlsAndNetwork(const bool mandatory_network_balance,
+                                              const double dt,
+                                              DeferredLogger& local_deferredLogger);
+
+            /// Update rank's notion of intersecting wells and their
+            /// associate solution variables.
+            ///
+            /// \param[in] reportStepIdx Report step.
+            ///
+            /// \param[in] enableWellPIScaling Whether or not to enable WELPI
+            ///   scaling.  Typically enabled (i.e., true) only at the start
+            ///   of a report step.
+            void initializeLocalWellStructure(const int  reportStepIdx,
+                                              const bool enableWellPIScaling);
+
+            /// Initialize group control modes/constraints and group solution state.
+            ///
+            /// \param[in] reportStepIdx Report step.
+            void initializeGroupStructure(const int reportStepIdx);
 
             // called at the end of a time step
-            void timeStepSucceeded(const double& simulationTime, const double dt);
+            void timeStepSucceeded(const double simulationTime, const double dt);
 
             // called at the end of a report step
             void endReportStep();
@@ -399,9 +469,23 @@ namespace Opm {
             // xw to update Well State
             void recoverWellSolutionAndUpdateWellState(const BVector& x);
 
+            // using the solution x to recover the solution xw for wells and applying
+            // xw to update Well State
+            void recoverWellSolutionAndUpdateWellStateDomain(const BVector& x, const Domain& domain);
+
             // setting the well_solutions_ based on well_state.
             void updatePrimaryVariables(DeferredLogger& deferred_logger);
 
+            void initializeWBPCalculationService();
+
+            data::WellBlockAveragePressures
+            computeWellBlockAveragePressures() const;
+
+            ParallelWBPCalculation::EvaluatorFactory
+            makeWellSourceEvaluatorFactory(const std::vector<Well>::size_type wellIdx) const;
+
+            void registerOpenWellsForWBPCalculation();
+
             void updateAverageFormationFactor();
 
             void computePotentials(const std::size_t widx,
@@ -423,6 +507,12 @@ namespace Opm {
             int reportStepIndex() const;
 
             void assembleWellEq(const double dt, DeferredLogger& deferred_logger);
+            void assembleWellEqDomain(const double dt, const Domain& domain, DeferredLogger& deferred_logger);
+
+            void prepareWellsBeforeAssembling(const double dt, DeferredLogger& deferred_logger);
+
+            // TODO: finding a better naming
+            void assembleWellEqWithoutIteration(const double dt, DeferredLogger& deferred_logger);
 
             bool maybeDoGasLiftOptimize(DeferredLogger& deferred_logger);
 
@@ -457,14 +547,12 @@ namespace Opm {
 
             void computeWellTemperature();
 
-            void assignWellTracerRates(data::Wells& wsrpt) const;
-
             int compressedIndexForInterior(int cartesian_cell_idx) const override {
-                return ebosSimulator_.vanguard().compressedIndexForInterior(cartesian_cell_idx);
+                return simulator_.vanguard().compressedIndexForInterior(cartesian_cell_idx);
             }
 
         private:
-            BlackoilWellModel(Simulator& ebosSimulator, const PhaseUsage& pu);
+            BlackoilWellModel(Simulator& simulator, const PhaseUsage& pu);
         };
 
 
