@@ -311,13 +311,13 @@ checkGroupProductionConstraints(const Group& group,
             double current_rate = 0.0;
             current_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
                                                               well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Aqua], true);
+                                                              pu.phase_pos[BlackoilPhases::Aqua], false);
             current_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
                                                               well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Liquid], true);
+                                                              pu.phase_pos[BlackoilPhases::Liquid], false);
             current_rate += WellGroupHelpers::sumWellResRates(group, wellModel_.schedule(),
                                                               well_state, reportStepIdx,
-                                                              pu.phase_pos[BlackoilPhases::Vapour], true);
+                                                              pu.phase_pos[BlackoilPhases::Vapour], false);
 
             // sum over all nodes
             current_rate = wellModel_.comm().sum(current_rate);
@@ -400,38 +400,75 @@ actionOnBrokenConstraints(const Group& group,
     }
 }
 
-void BlackoilWellModelConstraints::
+bool BlackoilWellModelConstraints::
 actionOnBrokenConstraints(const Group& group,
-                          const Group::ExceedAction& exceed_action,
+                          const int reportStepIdx,
+                          const Group::GroupLimitAction group_limit_action,
                           const Group::ProductionCMode& newControl,
+                          const WellState& well_state,
+                          std::optional<std::string>& worst_offending_well,
                           GroupState& group_state,
                           DeferredLogger& deferred_logger) const
 {
-    const Group::ProductionCMode oldControl = wellModel_.groupState().production_control(group.name());
+
+    bool changed = false;
+    const Group::ProductionCMode oldControl =
+        wellModel_.groupState().production_control(group.name());
 
     std::string ss;
-    switch(exceed_action) {
+    switch(group_limit_action.allRates) {
     case Group::ExceedAction::NONE: {
         if (oldControl != newControl && oldControl != Group::ProductionCMode::NONE) {
-            ss = fmt::format("Group production exceed action is NONE for group {}. Nothing happens.",
-                             group.name());
+            if ((group_limit_action.water == Group::ExceedAction::RATE &&
+                 newControl == Group::ProductionCMode::WRAT) ||
+                (group_limit_action.gas == Group::ExceedAction::RATE &&
+                 newControl == Group::ProductionCMode::GRAT) ||
+                (group_limit_action.liquid == Group::ExceedAction::RATE &&
+                 newControl == Group::ProductionCMode::LRAT)) {
+                group_state.production_control(group.name(), newControl);
+                ss = fmt::format("Switching production control mode for group {} from {} to {}",
+                                 group.name(),
+                                 Group::ProductionCMode2String(oldControl),
+                                 Group::ProductionCMode2String(newControl));
+
+                changed = true;
+            }
+            else {
+                ss = fmt::format("Procedure on exceeding {} limit is NONE for group {}. "
+                                 "Nothing is done.",
+                                 Group::ProductionCMode2String(oldControl),
+                                 group.name());
+            }
         }
         break;
     }
     case Group::ExceedAction::CON: {
-        OPM_DEFLOG_THROW(std::runtime_error, "Group " + group.name() + "GroupProductionExceedLimit CON not implemented", deferred_logger);
+        OPM_DEFLOG_THROW(std::runtime_error,
+                         fmt::format("Group {} GroupProductionExceedLimit CON not implemented",
+                                     group.name()),
+                         deferred_logger);
         break;
     }
     case Group::ExceedAction::CON_PLUS: {
-        OPM_DEFLOG_THROW(std::runtime_error, "Group " + group.name() + "GroupProductionExceedLimit CON_PLUS not implemented", deferred_logger);
+        OPM_DEFLOG_THROW(std::runtime_error,
+                         fmt::format("Group {} GroupProductionExceedLimit CON_PLUS not implemented",
+                                     group.name()),
+                         deferred_logger);
         break;
     }
     case Group::ExceedAction::WELL: {
-        OPM_DEFLOG_THROW(std::runtime_error, "Group " + group.name() + "GroupProductionExceedLimit WELL not implemented", deferred_logger);
+
+        std::tie(worst_offending_well, std::ignore) =
+            WellGroupHelpers::worstOffendingWell(group, wellModel_.schedule(), reportStepIdx,
+                                                 newControl, wellModel_.phaseUsage(),
+                                                 wellModel_.comm(), well_state, deferred_logger);
         break;
     }
     case Group::ExceedAction::PLUG: {
-        OPM_DEFLOG_THROW(std::runtime_error, "Group " + group.name() + "GroupProductionExceedLimit PLUG not implemented", deferred_logger);
+        OPM_DEFLOG_THROW(std::runtime_error,
+                         fmt::format("Group {} GroupProductionExceedLimit PLUG not implemented",
+                                     group.name()),
+                         deferred_logger);
         break;
     }
     case Group::ExceedAction::RATE: {
@@ -442,14 +479,18 @@ actionOnBrokenConstraints(const Group& group,
                              Group::ProductionCMode2String(oldControl),
                              Group::ProductionCMode2String(newControl));
         }
+        changed = true;
         break;
     }
     default:
-        throw("Invalid procedure for maximum rate limit selected for group" + group.name());
+        OPM_THROW(std::runtime_error,
+                  "Invalid procedure for maximum rate limit selected for group" + group.name());
     }
 
     if (!ss.empty() && wellModel_.comm().rank() == 0)
         deferred_logger.debug(ss);
+
+    return changed;
 }
 
 bool BlackoilWellModelConstraints::
@@ -457,6 +498,7 @@ updateGroupIndividualControl(const Group& group,
                              const int reportStepIdx,
                              std::map<std::pair<std::string,Opm::Phase>,std::string>& switched_inj,
                              std::map<std::string, std::string>& switched_prod,
+                             std::map<std::string, std::pair<std::string, std::string>>& closed_offending_wells,
                              GroupState& group_state,
                              WellState& well_state,
                              DeferredLogger& deferred_logger) const
@@ -494,23 +536,30 @@ updateGroupIndividualControl(const Group& group,
                                                                          reportStepIdx,
                                                                          deferred_logger);
         const auto controls = group.productionControls(wellModel_.summaryState());
+
         if (changed_this.first != Group::ProductionCMode::NONE)
         {
-            switched_prod.insert_or_assign(group.name(),
-                                  Group::ProductionCMode2String(changed_this.first));
-
-            this->actionOnBrokenConstraints(group,
-                                            controls.exceed_action,
-                                            changed_this.first,
+            std::optional<std::string> worst_offending_well = std::nullopt;
+            changed = this->actionOnBrokenConstraints(group, reportStepIdx,              
+                                            controls.group_limit_action,
+                                            changed_this.first, well_state, 
+                                            worst_offending_well,
                                             group_state, deferred_logger);
-            WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second,
-                                                                  group,
-                                                                  wellModel_.schedule(),
-                                                                  reportStepIdx,
-                                                                  /* isInjector */ false,
-                                                                  wellModel_.groupState(),
-                                                                  well_state);
-            changed = true;
+
+            if(changed) {
+                switched_prod.insert_or_assign(group.name(),
+                                    Group::ProductionCMode2String(changed_this.first));
+                WellGroupHelpers::updateWellRatesFromGroupTargetScale(changed_this.second,
+                                                                    group,
+                                                                    wellModel_.schedule(),
+                                                                    reportStepIdx,
+                                                                    /* isInjector */ false,
+                                                                    wellModel_.groupState(),
+                                                                    well_state);
+            } else if (worst_offending_well) {
+                closed_offending_wells.insert_or_assign(group.name(), 
+                            std::make_pair(Group::ProductionCMode2String(changed_this.first), *worst_offending_well));
+            }
         }
     }
 
