@@ -25,19 +25,28 @@
 
 #include <opm/output/data/GuideRateValue.hpp>
 
-#include <opm/input/eclipse/Schedule/Well/WellTestState.hpp>
 #include <opm/input/eclipse/Schedule/Group/GuideRate.hpp>
+#include <opm/input/eclipse/Schedule/Well/PAvg.hpp>
+#include <opm/input/eclipse/Schedule/Well/PAvgCalculator.hpp>
+#include <opm/input/eclipse/Schedule/Well/PAvgCalculatorCollection.hpp>
+#include <opm/input/eclipse/Schedule/Well/WellTestState.hpp>
 
 #include <opm/simulators/utils/DeferredLoggingErrorHelpers.hpp>
 
+#include <opm/simulators/wells/ParallelPAvgDynamicSourceData.hpp>
+#include <opm/simulators/wells/ParallelWBPCalculation.hpp>
 #include <opm/simulators/wells/PerforationData.hpp>
+#include <opm/simulators/wells/WellFilterCake.hpp>
 #include <opm/simulators/wells/WellProdIndexCalculator.hpp>
 #include <opm/simulators/wells/WGState.hpp>
 
+#include <cstddef>
 #include <functional>
 #include <map>
 #include <memory>
+#include <optional>
 #include <string>
+#include <type_traits>
 #include <unordered_map>
 #include <unordered_set>
 #include <vector>
@@ -45,14 +54,15 @@
 namespace Opm {
     class DeferredLogger;
     class EclipseState;
+    class GasLiftGroupInfo;
     class GasLiftSingleWellGeneric;
     class GasLiftWellState;
-    class GasLiftGroupInfo;
     class Group;
     class GuideRateConfig;
     class ParallelWellInfo;
     class RestartValue;
     class Schedule;
+    struct SimulatorUpdate;
     class SummaryConfig;
     class VFPProperties;
     class WellInterfaceGeneric;
@@ -86,12 +96,17 @@ public:
     virtual ~BlackoilWellModelGeneric() = default;
 
     int numLocalWells() const;
+    int numLocalWellsEnd() const;
+    int numLocalNonshutWells() const;
     int numPhases() const;
 
     /// return true if wells are available in the reservoir
     bool wellsActive() const;
     bool hasWell(const std::string& wname) const;
 
+    /// return true if network is active (at least one network well in prediction mode)
+    bool networkActive() const;
+
     // whether there exists any multisegment well open on this process
     bool anyMSWellOpenLocal() const;
 
@@ -119,6 +134,14 @@ public:
         return this->active_wgstate_.well_state;
     }
 
+    /*
+      Will return the currently active nupcolWellState; must initialize
+      the internal nupcol wellstate with initNupcolWellState() first.
+    */
+    const WellState& nupcolWellState() const
+    {
+        return this->nupcol_wgstate_.well_state;
+    }
     GroupState& groupState() { return this->active_wgstate_.group_state; }
 
     WellTestState& wellTestState() { return this->active_wgstate_.well_test_state; }
@@ -130,16 +153,16 @@ public:
     double wellPI(const std::string& well_name) const;
 
     void updateEclWells(const int timeStepIdx,
-                        const std::unordered_set<std::string>& wells,
+                        const SimulatorUpdate& sim_update,
                         const SummaryState& st);
 
     void initFromRestartFile(const RestartValue& restartValues,
                              WellTestState wtestState,
-                             const size_t numCells,
+                             const std::size_t numCells,
                              bool handle_ms_well);
 
     void prepareDeserialize(int report_step,
-                            const size_t numCells,
+                            const std::size_t numCells,
                             bool handle_ms_well);
 
     /*
@@ -158,6 +181,14 @@ public:
     /// Return true if any well has a THP constraint.
     bool hasTHPConstraints() const;
 
+    /// Checks if network is active (at least one network well on prediction).
+    void updateNetworkActiveState(const int report_step);
+
+    /// Checks if there are reasons to perform a pre-step network re-balance.
+    /// (Currently, the only reasons are network well status changes.)
+    /// (TODO: Consider if adding network change events would be helpful.)
+    bool needPreStepNetworkRebalance(const int report_step) const;
+
     /// Shut down any single well
     /// Returns true if the well was actually found and shut.
     bool forceShutWellByName(const std::string& wellname,
@@ -168,6 +199,8 @@ public:
 
     const Parallel::Communication& comm() const { return comm_; }
 
+    const EclipseState& eclipseState() const { return eclState_; }
+
     const SummaryState& summaryState() const { return summaryState_; }
 
     const GuideRate& guideRate() const { return guideRate_; }
@@ -177,9 +210,10 @@ public:
     bool shouldBalanceNetwork(const int reportStepIndex,
                               const int iterationIdx) const;
 
-    bool shouldIterateNetwork(const int reportStepIndex,
-                              const std::size_t recursion_level,
-                              const double network_imbalance) const;
+    void updateClosedWellsThisStep(const std::string& well_name) const {
+        this->closed_this_step_.insert(well_name);
+    }
+    bool wasDynamicallyShutThisTimeStep(const std::string& well_name) const;
 
     template<class Serializer>
     void serializeOp(Serializer& serializer)
@@ -191,12 +225,14 @@ public:
         serializer(closed_this_step_);
         serializer(guideRate_);
         serializer(node_pressures_);
+        serializer(prev_inj_multipliers_);
         serializer(active_wgstate_);
         serializer(last_valid_wgstate_);
         serializer(nupcol_wgstate_);
         serializer(last_glift_opt_time_);
         serializer(switched_prod_groups_);
         serializer(switched_inj_groups_);
+        serializer(closed_offending_wells_);
     }
 
     bool operator==(const BlackoilWellModelGeneric& rhs) const
@@ -207,12 +243,14 @@ public:
                this->local_shut_wells_ == rhs.local_shut_wells_ &&
                this->closed_this_step_ == rhs.closed_this_step_ &&
                this->node_pressures_ == rhs.node_pressures_ &&
+               this->prev_inj_multipliers_ == rhs.prev_inj_multipliers_ &&
                this->active_wgstate_ == rhs.active_wgstate_ &&
                this->last_valid_wgstate_ == rhs.last_valid_wgstate_ &&
                this->nupcol_wgstate_ == rhs.nupcol_wgstate_ &&
                this->last_glift_opt_time_ == rhs.last_glift_opt_time_ &&
                this->switched_prod_groups_ == rhs.switched_prod_groups_ &&
-               this->switched_inj_groups_ == rhs.switched_inj_groups_;
+               this->switched_inj_groups_ == rhs.switched_inj_groups_ &&
+               this->closed_offending_wells_ == rhs.closed_offending_wells_;            
     }
 
 protected:
@@ -250,18 +288,13 @@ protected:
         return this->last_valid_wgstate_.well_state;
     }
 
+
     const WGState& prevWGState() const
     {
         return this->last_valid_wgstate_;
     }
-    /*
-      Will return the currently active nupcolWellState; must initialize
-      the internal nupcol wellstate with initNupcolWellState() first.
-    */
-    const WellState& nupcolWellState() const
-    {
-        return this->nupcol_wgstate_.well_state;
-    }
+
+
 
     /*
       Will store a copy of the input argument well_state in the
@@ -302,7 +335,7 @@ protected:
 
     bool wasDynamicallyShutThisTimeStep(const int well_index) const;
 
-    std::pair<bool, double> updateNetworkPressures(const int reportStepIdx);
+    double updateNetworkPressures(const int reportStepIdx);
 
     void updateWsolvent(const Group& group,
                         const int reportStepIdx,
@@ -324,7 +357,7 @@ protected:
                             data::GroupData& gdata) const;
     void assignGroupValues(const int reportStepIdx,
                            std::map<std::string, data::GroupData>& gvalues) const;
-    void assignNodeValues(std::map<std::string, data::NodeData>& nodevalues) const;
+    void assignNodeValues(std::map<std::string, data::NodeData>& nodevalues, const int reportStepIdx) const;
 
     void calculateEfficiencyFactors(const int reportStepIdx);
 
@@ -333,6 +366,11 @@ protected:
                              const int reportStepIdx,
                              DeferredLogger& deferred_logger);
 
+    void checkGEconLimits(const Group& group,
+                          const double simulation_time,
+                          const int report_step_idx,
+                          DeferredLogger& deferred_logger);
+
     bool checkGroupHigherConstraints(const Group& group,
                                      DeferredLogger& deferred_logger,
                                      const int reportStepIdx);
@@ -368,6 +406,14 @@ protected:
                               const SummaryConfig& summaryConfig,
                               DeferredLogger& deferred_logger);
 
+    void initInjMult();
+
+
+    void updateInjMult(DeferredLogger& deferred_logger);
+    void updateInjFCMult(DeferredLogger& deferred_logger);
+
+    void updateFiltrationParticleVolume(const double dt, const std::size_t water_index);
+
     // create the well container
     virtual void createWellContainer(const int time_step) = 0;
     virtual void initWellContainer(const int reportStepIdx) = 0;
@@ -376,17 +422,22 @@ protected:
                                                            DeferredLogger& deferred_logger) = 0;
     virtual void calculateProductivityIndexValues(DeferredLogger& deferred_logger) = 0;
 
-    void runWellPIScaling(const int timeStepIdx,
+    void runWellPIScaling(const int reportStepIdx,
                           DeferredLogger& local_deferredLogger);
 
     /// \brief get compressed index for interior cells (-1, otherwise
     virtual int compressedIndexForInterior(int cartesian_cell_idx) const = 0;
 
     std::vector<int> getCellsForConnections(const Well& well) const;
+    std::vector<std::vector<int>> getMaxWellConnections() const;
 
     std::vector<std::string> getWellsForTesting(const int timeStepIdx,
                                                 const double simulationTime);
 
+    using WellTracerRates = std::map<std::pair<std::string, std::string>, double>;
+    void assignWellTracerRates(data::Wells& wsrpt,
+                               const WellTracerRates& wellTracerRates) const;
+
     Schedule& schedule_;
     const SummaryState& summaryState_;
     const EclipseState& eclState_;
@@ -395,6 +446,7 @@ protected:
     PhaseUsage phase_usage_;
     bool terminal_output_{false};
     bool wells_active_{false};
+    bool network_active_{false};
     bool initial_step_{};
     bool report_step_starts_{};
 
@@ -402,6 +454,94 @@ protected:
 
     std::vector<Well> wells_ecl_;
     std::vector<std::vector<PerforationData>> well_perf_data_;
+
+    /// Connection index mappings
+    class ConnectionIndexMap
+    {
+    public:
+        /// Constructor.
+        ///
+        /// \param[in] numConns Total number of well connections, both open
+        ///   and closed/shut.  Typically \code WellConnections::size() \endcode.
+        explicit ConnectionIndexMap(const std::size_t numConns)
+            : local_(numConns, -1)
+        {
+            this->global_.reserve(numConns);
+            this->open_.reserve(numConns);
+        }
+
+        /// Enumerate/map new active connection.
+        ///
+        /// \param[in] connIdx Global well connection index.  Must be an
+        ///   integer in the range 0..numConns-1.
+        ///
+        /// \param[in] connIsOpen Whether or not the connection is
+        ///   open/flowing.
+        void addActiveConnection(const int  connIdx,
+                                 const bool connIsOpen)
+        {
+            this->local_[connIdx] =
+                static_cast<int>(this->global_.size());
+
+            this->global_.push_back(connIdx);
+
+            const auto open_conn_idx = connIsOpen
+                ? this->num_open_conns_++
+                : -1;
+
+            this->open_.push_back(open_conn_idx);
+        }
+
+        /// Get local connection IDs/indices of every existing well
+        /// connection.
+        ///
+        /// Negative value (-1) for connections that don't intersect the
+        /// current rank.
+        const std::vector<int>& local() const
+        {
+            return this->local_;
+        }
+
+        /// Get global connection ID of local (on-rank) connection.
+        ///
+        /// \param[in] connIdx Local connection index.
+        ///
+        /// \return Global connection ID of \p connIdx.
+        int global(const int connIdx) const
+        {
+            return this->global_[connIdx];
+        }
+
+        /// Get open connection ID of local (on-rank) connection.
+        ///
+        /// \param[in] connIdx Local connection index.
+        ///
+        /// \return Open connection ID of \p connIdx.  Integer in the range
+        ///   0..#open connections - 1 if the connection is open or negative
+        ///   value (-1) otherwise.
+        int open(const int connIdx) const
+        {
+            return this->open_[connIdx];
+        }
+
+    private:
+        /// Local connection IDs/indices of every existing well connection.
+        /// Negative value (-1) for connections that don't intersect the
+        /// current rank.
+        std::vector<int> local_{};
+
+        /// Global connection index of each on-rank reservoir connection.
+        /// Reverse/transpose mapping of \c local_.
+        std::vector<int> global_{};
+
+        /// Open connection index of each on-rank reservoir connection.
+        std::vector<int> open_{};
+
+        /// Number of open connections on this rank.
+        int num_open_conns_{0};
+    };
+
+    std::vector<ConnectionIndexMap> conn_idx_map_{};
     std::function<bool(const Well&)> not_on_process_{};
 
     // a vector of all the wells.
@@ -413,6 +553,7 @@ protected:
     std::vector<std::reference_wrapper<ParallelWellInfo>> local_parallel_well_info_;
 
     std::vector<WellProdIndexCalculator> prod_index_calc_;
+    mutable ParallelWBPCalculation wbpCalculationService_;
 
     std::vector<int> pvt_region_idx_;
 
@@ -422,6 +563,12 @@ protected:
     std::unique_ptr<VFPProperties> vfp_properties_{};
     std::map<std::string, double> node_pressures_; // Storing network pressures for output.
 
+    // previous injection multiplier, it is used in the injection multiplier calculation for WINJMULT keyword
+    std::unordered_map<std::string, std::vector<double>> prev_inj_multipliers_;
+
+    // Handling for filter cake injection multipliers
+    std::unordered_map<std::string, WellFilterCake> filter_cake_;
+
     /*
       The various wellState members should be accessed and modified
       through the accessor functions wellState(), prevWellState(),
@@ -436,8 +583,13 @@ protected:
 
     double last_glift_opt_time_ = -1.0;
 
+    bool wellStructureChangedDynamically_{false};
+
+    // Store maps of group name and new group controls for output
     std::map<std::string, std::string> switched_prod_groups_;
     std::map<std::pair<std::string, Opm::Phase>, std::string> switched_inj_groups_;
+    // Store map of group name and close offending well for output
+    std::map<std::string, std::pair<std::string, std::string>> closed_offending_wells_;
 
 
 private:
