@@ -51,9 +51,9 @@
 namespace Opm
 {
 
-template<typename FluidSystem, typename Indices, typename Scalar>
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
-MultisegmentWellEval(WellInterfaceIndices<FluidSystem,Indices,Scalar>& baseif)
+template<typename FluidSystem, typename Indices>
+MultisegmentWellEval<FluidSystem,Indices>::
+MultisegmentWellEval(WellInterfaceIndices<FluidSystem,Indices>& baseif)
     : MultisegmentWellGeneric<Scalar>(baseif)
     , baseif_(baseif)
     , linSys_(*this)
@@ -64,9 +64,9 @@ MultisegmentWellEval(WellInterfaceIndices<FluidSystem,Indices,Scalar>& baseif)
 {
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 void
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+MultisegmentWellEval<FluidSystem,Indices>::
 initMatrixAndVectors(const int num_cells)
 {
     linSys_.init(num_cells, baseif_.numPerfs(),
@@ -75,9 +75,9 @@ initMatrixAndVectors(const int num_cells)
     primary_variables_.resize(this->numberOfSegments());
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 ConvergenceReport
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+MultisegmentWellEval<FluidSystem,Indices>::
 getWellConvergence(const WellState& well_state,
                    const std::vector<double>& B_avg,
                    DeferredLogger& deferred_logger,
@@ -86,7 +86,8 @@ getWellConvergence(const WellState& well_state,
                    const double relaxed_inner_tolerance_flow_ms_well,
                    const double tolerance_pressure_ms_wells,
                    const double relaxed_inner_tolerance_pressure_ms_well,
-                   const bool relax_tolerance) const
+                   const bool relax_tolerance, 
+                   const bool well_is_stopped) const
 {
     assert(int(B_avg.size()) == baseif_.numComponents());
 
@@ -161,12 +162,13 @@ getWellConvergence(const WellState& well_state,
                                    tolerance_wells,
                                    max_residual_allowed},
                                   std::abs(linSys_.residual()[0][SPres]),
+                                  well_is_stopped,  
                                   report,
                                   deferred_logger);
 
     // for stopped well, we do not enforce the following checking to avoid dealing with sign of near-zero values
     // for BHP or THP controlled wells, we need to make sure the flow direction is correct
-    if (!baseif_.wellIsStopped() && baseif_.isPressureControlled(well_state)) {
+    if (!well_is_stopped && baseif_.isPressureControlled(well_state)) {
         // checking the flow direction
         const double sign = baseif_.isProducer() ? -1. : 1.;
         const auto weight_total_flux = this->primary_variables_.getWQTotal() * sign;
@@ -180,9 +182,9 @@ getWellConvergence(const WellState& well_state,
     return report;
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
-typename MultisegmentWellEval<FluidSystem,Indices,Scalar>::EvalWell
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+template<typename FluidSystem, typename Indices>
+typename MultisegmentWellEval<FluidSystem,Indices>::EvalWell
+MultisegmentWellEval<FluidSystem,Indices>::
 extendEval(const Eval& in) const
 {
     EvalWell out = 0.0;
@@ -193,46 +195,84 @@ extendEval(const Eval& in) const
     return out;
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 void
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
-handleAccelerationPressureLoss(const int seg,
-                               WellState& well_state)
+MultisegmentWellEval<FluidSystem,Indices>::
+assembleAccelerationPressureLoss(const int seg,
+                                 WellState& well_state)
 {
-    const EvalWell accelerationPressureLoss = segments_.accelerationPressureLoss(seg);
-
+    // Computes and assembles p-drop due to acceleration
+    assert(seg != 0); // top segment can not enter here
     auto& segments = well_state.well(baseif_.indexOfWell()).segments;
-    segments.pressure_drop_accel[seg] = accelerationPressureLoss.value();
+    const auto& segment_set = this->segmentSet();
 
-    MultisegmentWellAssemble<FluidSystem,Indices,Scalar>(baseif_).
-        assemblePressureLoss(seg,
-                             segments_.upwinding_segment(seg),
-                             accelerationPressureLoss, linSys_);
+    // Add segment head
+    const double seg_area = segment_set[seg].crossArea();
+    const EvalWell signed_velocity_head = segments_.accelerationPressureLossContribution(seg, seg_area);
+    segments.pressure_drop_accel[seg] = signed_velocity_head.value();
+    
+    const int seg_upwind = segments_.upwinding_segment(seg);
+    // acceleration term is *subtracted* from pressure equation
+    MultisegmentWellAssemble(baseif_).
+        assembleAccelerationTerm(seg, seg, seg_upwind, signed_velocity_head, linSys_);
+    if (seg != seg_upwind) {// special treatment for reverse flow
+        // extra derivatives are *added* to Jacobian (hence minus)
+        const EvalWell extra_derivatives = -segments_.accelerationPressureLossContribution(seg, seg_area, /*extra_derivatives*/ true);
+        MultisegmentWellAssemble(baseif_).
+            assemblePressureEqExtraDerivatives(seg, seg_upwind, extra_derivatives, linSys_);
+    }
+
+    // Subtract inlet head(s), opposite signs from above
+    for (const int inlet : segments_.inlets(seg)) {
+        // area used in formula is max of areas
+        const double inlet_area = std::max(seg_area, segment_set[inlet].crossArea());
+        const EvalWell signed_velocity_head_inlet = segments_.accelerationPressureLossContribution(inlet, inlet_area);
+        segments.pressure_drop_accel[seg] -= signed_velocity_head_inlet.value();
+
+        const int inlet_upwind = segments_.upwinding_segment(inlet); 
+        MultisegmentWellAssemble(baseif_).
+            assembleAccelerationTerm(seg, inlet, inlet_upwind, -signed_velocity_head_inlet, linSys_);
+        if (inlet != inlet_upwind) {// special treatment for reverse flow
+            // extra derivatives are *added* to Jacobian (hence minus minus)
+            const EvalWell extra_derivatives_inlet = segments_.accelerationPressureLossContribution(inlet, inlet_area, /*extra_derivatives*/ true);
+            // in this case inlet_upwind = seg
+            MultisegmentWellAssemble(baseif_).
+                assemblePressureEqExtraDerivatives(seg, inlet_upwind, extra_derivatives_inlet, linSys_);
+        }
+    }
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 void
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+MultisegmentWellEval<FluidSystem,Indices>::
 assembleDefaultPressureEq(const int seg,
-                          WellState& well_state)
+                          WellState& well_state,
+                          const bool use_average_density)
 {
     assert(seg != 0); // not top segment
+    const int seg_upwind = segments_.upwinding_segment(seg);
+    const bool reverseFlow = seg != seg_upwind; // special treatment for reverse flow
 
     // for top segment, the well control equation will be used.
     EvalWell pressure_equation = primary_variables_.getSegmentPressure(seg);
+    EvalWell extra_derivatives;
 
     // we need to handle the pressure difference between the two segments
-    // we only consider the hydrostatic pressure loss first
+    // hydrostatic pressure loss is assembled seperately at the end
     // TODO: we might be able to add member variables to store these values, then we update well state
     // after converged
-    const auto hydro_pressure_drop = segments_.getHydroPressureLoss(seg);
+
     auto& ws = well_state.well(baseif_.indexOfWell());
     auto& segments = ws.segments;
-    segments.pressure_drop_hydrostatic[seg] = hydro_pressure_drop.value();
-    pressure_equation -= hydro_pressure_drop;
 
     if (this->frictionalPressureLossConsidered()) {
         const auto friction_pressure_drop = segments_.getFrictionPressureLoss(seg);
+        if (reverseFlow){
+            // call function once again to obtain/assemble remaining derivatives
+            extra_derivatives = -segments_.getFrictionPressureLoss(seg, /*extra_reverse_flow_derivatives*/ true);
+            MultisegmentWellAssemble(baseif_).
+                assemblePressureEqExtraDerivatives(seg, seg_upwind, extra_derivatives, linSys_);
+        }
         pressure_equation -= friction_pressure_drop;
         segments.pressure_drop_friction[seg] = friction_pressure_drop.value();
     }
@@ -241,22 +281,21 @@ assembleDefaultPressureEq(const int seg,
     const int outlet_segment_index = this->segmentNumberToIndex(this->segmentSet()[seg].outletSegment());
     const EvalWell outlet_pressure = primary_variables_.getSegmentPressure(outlet_segment_index);
 
-    const int seg_upwind = segments_.upwinding_segment(seg);
-    MultisegmentWellAssemble<FluidSystem,Indices,Scalar>(baseif_).
+    MultisegmentWellAssemble(baseif_).
         assemblePressureEq(seg, seg_upwind, outlet_segment_index,
                            pressure_equation, outlet_pressure, linSys_);
 
-    if (this->accelerationalPressureLossConsidered()) {
-        handleAccelerationPressureLoss(seg, well_state);
-    }
+    assembleAccelerationAndHydroPressureLosses(seg, well_state, use_average_density);
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 void
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+MultisegmentWellEval<FluidSystem,Indices>::
 assembleICDPressureEq(const int seg,
                       const UnitSystem& unit_system,
                       WellState& well_state,
+                      const SummaryState& summary_state,
+                      const bool use_average_density,
                       DeferredLogger& deferred_logger)
 {
     // TODO: upwinding needs to be taken care of
@@ -266,7 +305,7 @@ assembleICDPressureEq(const int seg,
     if (const auto& segment = this->segmentSet()[seg];
        (segment.segmentType() == Segment::SegmentType::VALVE) &&
        (segment.valve().status() == Opm::ICDStatus::SHUT) ) { // we use a zero rate equation to handle SHUT valve
-        MultisegmentWellAssemble<FluidSystem,Indices,Scalar>(baseif_).
+        MultisegmentWellAssemble(baseif_).
             assembleTrivialEq(seg, this->primary_variables_.eval(seg)[WQTotal].value(), linSys_);
 
         auto& ws = well_state.well(baseif_.indexOfWell());
@@ -277,19 +316,31 @@ assembleICDPressureEq(const int seg,
     // the pressure equation is something like
     // p_seg - deltaP - p_outlet = 0.
     // the major part is how to calculate the deltaP
+    const int seg_upwind = segments_.upwinding_segment(seg);
+    const bool reverseFlow = seg != seg_upwind; // special treatment for reverse flow
 
     EvalWell pressure_equation = primary_variables_.getSegmentPressure(seg);
 
     EvalWell icd_pressure_drop;
+    EvalWell extra_derivatives;
     switch(this->segmentSet()[seg].segmentType()) {
         case Segment::SegmentType::SICD :
             icd_pressure_drop = segments_.pressureDropSpiralICD(seg);
+            if (reverseFlow){
+                extra_derivatives = -segments_.pressureDropSpiralICD(seg, /*extra_reverse_flow_derivatives*/ true);
+            }
             break;
         case Segment::SegmentType::AICD :
             icd_pressure_drop = segments_.pressureDropAutoICD(seg, unit_system);
+            if (reverseFlow){
+                extra_derivatives = -segments_.pressureDropAutoICD(seg, unit_system, /*extra_reverse_flow_derivatives*/ true);
+            }
             break;
         case Segment::SegmentType::VALVE :
-            icd_pressure_drop = segments_.pressureDropValve(seg);
+            icd_pressure_drop = segments_.pressureDropValve(seg, summary_state);
+            if (reverseFlow){
+                extra_derivatives = -segments_.pressureDropValve(seg, summary_state, /*extra_reverse_flow_derivatives*/ true);
+            }
             break;
         default: {
             OPM_DEFLOG_THROW(std::runtime_error,
@@ -299,6 +350,11 @@ assembleICDPressureEq(const int seg,
                              deferred_logger);
         }
     }
+    if (reverseFlow){
+        MultisegmentWellAssemble(baseif_).
+            assemblePressureEqExtraDerivatives(seg, seg_upwind, extra_derivatives, linSys_);
+    }
+
     pressure_equation = pressure_equation - icd_pressure_drop;
     auto& ws = well_state.well(baseif_.indexOfWell());
     ws.segments.pressure_drop_friction[seg] = icd_pressure_drop.value();
@@ -307,38 +363,72 @@ assembleICDPressureEq(const int seg,
     const int outlet_segment_index = this->segmentNumberToIndex(this->segmentSet()[seg].outletSegment());
     const EvalWell outlet_pressure = primary_variables_.getSegmentPressure(outlet_segment_index);
 
-    const int seg_upwind = segments_.upwinding_segment(seg);
-    MultisegmentWellAssemble<FluidSystem,Indices,Scalar>(baseif_).
+    MultisegmentWellAssemble(baseif_).
         assemblePressureEq(seg, seg_upwind, outlet_segment_index,
                            pressure_equation, outlet_pressure,
                            linSys_,
                            FluidSystem::phaseIsActive(FluidSystem::waterPhaseIdx),
                            FluidSystem::phaseIsActive(FluidSystem::gasPhaseIdx));
+
+    assembleAccelerationAndHydroPressureLosses(seg, well_state, use_average_density);
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 void
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+MultisegmentWellEval<FluidSystem,Indices>::
+assembleAccelerationAndHydroPressureLosses(const int seg,
+                                           WellState& well_state,
+                                           const bool use_average_density)
+{
+    if (this->accelerationalPressureLossConsidered()) {
+        assembleAccelerationPressureLoss(seg, well_state);
+    }
+
+    // Since density derivatives are organized differently than what is required for assemblePressureEq,
+    // this part needs to be assembled separately. Optionally use average density variant.
+    const auto hydro_pressure_drop_seg = segments_.getHydroPressureLoss(seg, seg);
+    auto& ws = well_state.well(baseif_.indexOfWell());
+    auto& segments = ws.segments;
+    if (!use_average_density){
+        MultisegmentWellAssemble(baseif_).
+            assembleHydroPressureLoss(seg, seg, hydro_pressure_drop_seg, linSys_);
+        segments.pressure_drop_hydrostatic[seg] = hydro_pressure_drop_seg.value();
+    } else {
+        const int seg_outlet = this->segmentNumberToIndex(this->segmentSet()[seg].outletSegment());
+        const auto hydro_pressure_drop_outlet = segments_.getHydroPressureLoss(seg, seg_outlet);
+        MultisegmentWellAssemble(baseif_).
+            assembleHydroPressureLoss(seg, seg, 0.5*hydro_pressure_drop_seg, linSys_);
+        MultisegmentWellAssemble(baseif_).
+            assembleHydroPressureLoss(seg, seg_outlet, 0.5*hydro_pressure_drop_outlet, linSys_);
+        segments.pressure_drop_hydrostatic[seg] = 0.5*hydro_pressure_drop_seg.value() + 0.5*hydro_pressure_drop_outlet.value();
+    }
+}
+
+template<typename FluidSystem, typename Indices>
+void
+MultisegmentWellEval<FluidSystem,Indices>::
 assemblePressureEq(const int seg,
                    const UnitSystem& unit_system,
                    WellState& well_state,
+                   const SummaryState& summary_state,
+                   const bool use_average_density,
                    DeferredLogger& deferred_logger)
 {
     switch(this->segmentSet()[seg].segmentType()) {
         case Segment::SegmentType::SICD :
         case Segment::SegmentType::AICD :
         case Segment::SegmentType::VALVE : {
-            assembleICDPressureEq(seg, unit_system, well_state,deferred_logger);
+            assembleICDPressureEq(seg, unit_system, well_state, summary_state, use_average_density, deferred_logger);
             break;
         }
         default :
-            assembleDefaultPressureEq(seg, well_state);
+            assembleDefaultPressureEq(seg, well_state, use_average_density);
     }
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
-std::pair<bool, std::vector<Scalar> >
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+template<typename FluidSystem, typename Indices>
+std::pair<bool, std::vector<typename FluidSystem::Scalar> >
+MultisegmentWellEval<FluidSystem,Indices>::
 getFiniteWellResiduals(const std::vector<Scalar>& B_avg,
                        DeferredLogger& deferred_logger) const
 {
@@ -380,9 +470,9 @@ getFiniteWellResiduals(const std::vector<Scalar>& B_avg,
     return {true, residuals};
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 double
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+MultisegmentWellEval<FluidSystem,Indices>::
 getControlTolerance(const WellState& well_state,
                     const double tolerance_wells,
                     const double tolerance_pressure_ms_wells,
@@ -447,9 +537,9 @@ getControlTolerance(const WellState& well_state,
     return control_tolerance;
 }
 
-template<typename FluidSystem, typename Indices, typename Scalar>
+template<typename FluidSystem, typename Indices>
 double
-MultisegmentWellEval<FluidSystem,Indices,Scalar>::
+MultisegmentWellEval<FluidSystem,Indices>::
 getResidualMeasureValue(const WellState& well_state,
                         const std::vector<double>& residuals,
                         const double tolerance_wells,
@@ -459,7 +549,7 @@ getResidualMeasureValue(const WellState& well_state,
     assert(int(residuals.size()) == numWellEq + 1);
 
     const double rate_tolerance = tolerance_wells;
-    int count = 0;
+    [[maybe_unused]] int count = 0;
     double sum = 0;
     for (int eq_idx = 0; eq_idx < numWellEq - 1; ++eq_idx) {
         if (residuals[eq_idx] > rate_tolerance) {
@@ -483,14 +573,11 @@ getResidualMeasureValue(const WellState& well_state,
         ++count;
     }
 
-    // if (count == 0), it should be converged.
-    assert(count != 0);
-
     return sum;
 }
 
 #define INSTANCE(...) \
-template class MultisegmentWellEval<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,__VA_ARGS__,double>;
+template class MultisegmentWellEval<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,__VA_ARGS__>;
 
 // One phase
 INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
@@ -507,6 +594,9 @@ INSTANCE(BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>)
 INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,1u,0u>)
 INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)
 INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>)
+INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,true,0u,0u,0u>)
+INSTANCE(BlackOilTwoPhaseIndices<1u,0u,0u,0u,false,false,0u,0u,0u>)
+
 // Blackoil
 INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)
 INSTANCE(BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)
@@ -518,5 +608,6 @@ INSTANCE(BlackOilIndices<0u,0u,1u,0u,false,false,0u,0u>)
 INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,false,0u,0u>)
 INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,1u,0u>)
 INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)
+INSTANCE(BlackOilIndices<1u,0u,0u,0u,true,false,0u,0u>)
 
 } // namespace Opm
