@@ -1,7 +1,6 @@
 /*
   Copyright 2021 Equinor ASA.
 
-
   This file is part of the Open Porous Media project (OPM).
 
   OPM is free software: you can redistribute it and/or modify
@@ -21,8 +20,11 @@
 #ifndef OPM_PERFDATA_HEADER_INCLUDED
 #define OPM_PERFDATA_HEADER_INCLUDED
 
+#include <opm/simulators/wells/ConnFiltrateData.hpp>
+
 #include <cstddef>
 #include <vector>
+#include <array>
 
 namespace Opm {
 
@@ -33,7 +35,10 @@ private:
 
 public:
     PerfData() = default;
-    PerfData(std::size_t num_perf, double pressure_first_connection_, bool injector_, std::size_t num_phases);
+    PerfData(std::size_t num_perf,
+             double pressure_first_connection_,
+             bool injector_,
+             std::size_t num_phases);
 
     static PerfData serializationTestObject();
 
@@ -48,6 +53,7 @@ public:
         serializer(pressure);
         serializer(rates);
         serializer(phase_rates);
+        serializer(phase_mixing_rates);
         serializer(solvent_rates);
         serializer(polymer_rates);
         serializer(brine_rates);
@@ -55,37 +61,50 @@ public:
         serializer(micp_rates);
         serializer(cell_index);
         serializer(connection_transmissibility_factor);
+        serializer(connection_d_factor);
+        serializer(connection_compaction_tmult);
         serializer(satnum_id);
         serializer(ecl_index);
         serializer(water_throughput);
         serializer(skin_pressure);
         serializer(water_velocity);
+        serializer(filtrate_data);
     }
 
     bool operator==(const PerfData&) const;
 
+    // Note to maintainers: If you make changes to this list of data
+    // members, then please update the constructor, operator==(),
+    // serializationTestObject(), and serializeOp() accordingly.  Moreover,
+    // if you're adding a new member representing a dynamically calculated
+    // result, e.g., a flow rate, then please update try_assign() as well.
+
     double pressure_first_connection{};
-    std::vector<double> pressure;
-    std::vector<double> rates;
-    std::vector<double> phase_rates;
-    std::vector<double> solvent_rates;
-    std::vector<double> polymer_rates;
-    std::vector<double> brine_rates;
-    std::vector<double> prod_index;
-    std::vector<double> micp_rates;
+    std::vector<double> pressure{};
+    std::vector<double> rates{};
+    std::vector<double> phase_rates{};
+    std::vector<std::array<double,4>> phase_mixing_rates{};
+    std::vector<double> solvent_rates{};
+    std::vector<double> polymer_rates{};
+    std::vector<double> brine_rates{};
+    std::vector<double> prod_index{};
+    std::vector<double> micp_rates{};
+    std::vector<std::size_t> cell_index{};
+    std::vector<double> connection_transmissibility_factor{};
+    std::vector<double> connection_d_factor{};
+    std::vector<double> connection_compaction_tmult{};
+    std::vector<int> satnum_id{};
+    std::vector<std::size_t> ecl_index{};
 
-    std::vector<std::size_t> cell_index;
-    std::vector<double> connection_transmissibility_factor;
-    std::vector<int> satnum_id;
-    std::vector<std::size_t> ecl_index;
+    // The water_throughput, skin_pressure and water_velocity variables are
+    // only used for injectors to check the injectivity.
+    std::vector<double> water_throughput{};
+    std::vector<double> skin_pressure{};
+    std::vector<double> water_velocity{};
 
-    // The water_throughput, skin_pressure and water_velocity variables are only
-    // used for injectors to check the injectivity.
-    std::vector<double> water_throughput;
-    std::vector<double> skin_pressure;
-    std::vector<double> water_velocity;
+    ConnFiltrateData filtrate_data{};
 };
 
 } // namespace Opm
 
-#endif // OPM_PERFORATIONDATA_HEADER_INCLUDED
+#endif // OPM_PERFDATA_HEADER_INCLUDED
