@@ -58,19 +58,20 @@ public:
         serializer(bhp);
         serializer(thp);
         serializer(temperature);
-        serializer(dissolved_gas_rate);
-        serializer(dissolved_gas_rate_in_water);
-        serializer(vaporized_oil_rate);
-        serializer(vaporized_wat_rate);
+        serializer(phase_mixing_rates);
         serializer(well_potentials);
         serializer(productivity_index);
+        serializer(implicit_ipr_a);
+        serializer(implicit_ipr_b);        
         serializer(surface_rates);
         serializer(reservoir_rates);
+        serializer(prev_surface_rates);
         serializer(trivial_target);
         serializer(segments);
         serializer(events);
         serializer(injection_cmode);
         serializer(production_cmode);
+        serializer(filtrate_conc);
         serializer(perf_data);
     }
 
@@ -85,14 +86,25 @@ public:
     double bhp{0};
     double thp{0};
     double temperature{0};
-    double dissolved_gas_rate{0};
-    double dissolved_gas_rate_in_water{0};
-    double vaporized_oil_rate{0};
-    double vaporized_wat_rate{0};
+
+    // filtration injection concentration
+    double filtrate_conc{0};
+
+    std::array<double,4> phase_mixing_rates{};
+    enum RateIndices {
+      dissolved_gas = 0,
+      dissolved_gas_in_water = 1,
+      vaporized_oil = 2,
+      vaporized_water = 3
+    };
+
     std::vector<double> well_potentials;
     std::vector<double> productivity_index;
+    std::vector<double> implicit_ipr_a;
+    std::vector<double> implicit_ipr_b;    
     std::vector<double> surface_rates;
     std::vector<double> reservoir_rates;
+    std::vector<double> prev_surface_rates;
     PerfData perf_data;
     bool trivial_target;
     SegmentState segments;
@@ -124,6 +136,9 @@ public:
     double sum_polymer_rates() const;
     double sum_brine_rates() const;
 
+    double sum_filtrate_rate() const;
+    double sum_filtrate_total() const;
+
 private:
     double sum_connection_rates(const std::vector<double>& connection_rates) const;
 };
