@@ -55,15 +55,13 @@ namespace Opm
     template<typename TypeTag>
     class StandardWell : public WellInterface<TypeTag>
                        , public StandardWellEval<GetPropType<TypeTag, Properties::FluidSystem>,
-                                                 GetPropType<TypeTag, Properties::Indices>,
-                                                 GetPropType<TypeTag, Properties::Scalar>>
+                                                 GetPropType<TypeTag, Properties::Indices>>
     {
 
     public:
         using Base = WellInterface<TypeTag>;
         using StdWellEval = StandardWellEval<GetPropType<TypeTag, Properties::FluidSystem>,
-                                             GetPropType<TypeTag, Properties::Indices>,
-                                             GetPropType<TypeTag, Properties::Scalar>>;
+                                             GetPropType<TypeTag, Properties::Indices>>;
 
         // TODO: some functions working with AD variables handles only with values (double) without
         // dealing with derivatives. It can be beneficial to make functions can work with either AD or scalar value.
@@ -144,10 +142,11 @@ namespace Opm
         void initPrimaryVariablesEvaluation() override;
 
         /// check whether the well equations get converged for this well
-        virtual ConvergenceReport getWellConvergence(const WellState& well_state,
+        virtual ConvergenceReport getWellConvergence(const SummaryState& summary_state,
+                                                     const WellState& well_state,
                                                      const std::vector<double>& B_avg,
                                                      DeferredLogger& deferred_logger,
-                                                     const bool relax_tolerance = false) const override;
+                                                     const bool relax_tolerance) const override;
 
         /// Ax = Ax - C D^-1 B x
         virtual void apply(const BVector& x, BVector& Ax) const override;
@@ -162,7 +161,7 @@ namespace Opm
                                                    DeferredLogger& deferred_logger) override;
 
         /// computing the well potentials for group control
-        virtual void computeWellPotentials(const Simulator& ebosSimulator,
+        virtual void computeWellPotentials(const Simulator& simulator,
                                            const WellState& well_state,
                                            std::vector<double>& well_potentials,
                                            DeferredLogger& deferred_logger) /* const */ override;
@@ -171,19 +170,22 @@ namespace Opm
                                     const WellState& well_state,
                                     DeferredLogger& deferred_logger) override;
 
-        void solveEqAndUpdateWellState(const SummaryState& summary_state,
-                                       WellState& well_state,
-                                       DeferredLogger& deferred_logger);
+        virtual void solveEqAndUpdateWellState(const SummaryState& summary_state,
+                                               WellState& well_state,
+                                               DeferredLogger& deferred_logger) override;
 
-        virtual void calculateExplicitQuantities(const Simulator& ebosSimulator,
+        virtual void calculateExplicitQuantities(const Simulator& simulator,
                                                  const WellState& well_state,
                                                  DeferredLogger& deferred_logger) override; // should be const?
 
-        virtual void updateProductivityIndex(const Simulator& ebosSimulator,
+        virtual void updateProductivityIndex(const Simulator& simulator,
                                              const WellProdIndexCalculator& wellPICalc,
                                              WellState& well_state,
                                              DeferredLogger& deferred_logger) const override;
 
+        virtual double connectionDensity(const int globalConnIdx,
+                                         const int openConnIdx) const override;
+
         virtual void addWellContributions(SparseMatrixAdapter& mat) const override;
 
         virtual void addWellPressureEquations(PressureMatrix& mat,
@@ -193,7 +195,7 @@ namespace Opm
                                               const WellState& well_state) const override;
 
         // iterate well equations with the specified control until converged
-        bool iterateWellEqWithControl(const Simulator& ebosSimulator,
+        bool iterateWellEqWithControl(const Simulator& simulator,
                                       const double dt,
                                       const Well::InjectionControls& inj_controls,
                                       const Well::ProductionControls& prod_controls,
@@ -201,6 +203,17 @@ namespace Opm
                                       const GroupState& group_state,
                                       DeferredLogger& deferred_logger) override;
 
+        // iterate well equations including control switching
+        bool iterateWellEqWithSwitching(const Simulator& simulator,
+                                        const double dt,
+                                        const Well::InjectionControls& inj_controls,
+                                        const Well::ProductionControls& prod_controls,
+                                        WellState& well_state,
+                                        const GroupState& group_state,
+                                        DeferredLogger& deferred_logger, 
+                                        const bool fixed_control = false,
+                                        const bool fixed_status = false) override;
+
         /// \brief Wether the Jacobian will also have well contributions in it.
         virtual bool jacobianContainsWellContributions() const override
         {
@@ -208,27 +221,31 @@ namespace Opm
         }
 
         /* returns BHP */
-        double computeWellRatesAndBhpWithThpAlqProd(const Simulator &ebos_simulator,
-                               const SummaryState &summary_state,
-                               DeferredLogger &deferred_logger,
-                               std::vector<double> &potentials,
+        double computeWellRatesAndBhpWithThpAlqProd(const Simulator& simulator,
+                               const SummaryState& summary_state,
+                               DeferredLogger& deferred_logger,
+                               std::vector<double>& potentials,
                                double alq) const;
 
         void computeWellRatesWithThpAlqProd(
-            const Simulator &ebos_simulator,
-            const SummaryState &summary_state,
-            DeferredLogger &deferred_logger,
-            std::vector<double> &potentials,
+            const Simulator& simulator,
+            const SummaryState& summary_state,
+            DeferredLogger& deferred_logger,
+            std::vector<double>& potentials,
             double alq) const;
 
         std::optional<double> computeBhpAtThpLimitProdWithAlq(
-            const Simulator& ebos_simulator,
+            const Simulator& simulator,
             const SummaryState& summary_state,
             const double alq_value,
             DeferredLogger& deferred_logger) const override;
 
+        void updateIPRImplicit(const Simulator& simulator,
+                               WellState& well_state,
+                               DeferredLogger& deferred_logger) override;
+
         virtual void computeWellRatesWithBhp(
-            const Simulator& ebosSimulator,
+            const Simulator& simulator,
             const double& bhp,
             std::vector<double>& well_flux,
             DeferredLogger& deferred_logger) const override;
@@ -237,21 +254,12 @@ namespace Opm
         using Base::phaseUsage;
         using Base::vfp_properties_;
 
-        virtual std::vector<double> computeCurrentWellRates(const Simulator& ebosSimulator,
-                                                            DeferredLogger& deferred_logger) const override;
+        std::vector<double> computeCurrentWellRates(const Simulator& simulator,
+                                                    DeferredLogger& deferred_logger) const override;
 
-        void computeConnLevelProdInd(const FluidState& fs,
-                                     const std::function<double(const double)>& connPICalc,
-                                     const std::vector<EvalWell>& mobility,
-                                     double* connPI) const;
-
-        void computeConnLevelInjInd(const typename StandardWell<TypeTag>::FluidState& fs,
-                                    const Phase preferred_phase,
-                                    const std::function<double(const double)>& connIICalc,
-                                    const std::vector<EvalWell>& mobility,
-                                    double* connII,
-                                    DeferredLogger& deferred_logger) const;
+        std::vector<double> getPrimaryVars() const override;
 
+        int setPrimaryVars(std::vector<double>::const_iterator it) override;
 
     protected:
         bool regularize_;
@@ -264,50 +272,30 @@ namespace Opm
 
         // calculate the properties for the well connections
         // to calulate the pressure difference between well connections.
-        void computePropertiesForWellConnectionPressures(const Simulator& ebosSimulator,
+        using WellConnectionProps = typename StdWellEval::StdWellConnections::Properties;
+        void computePropertiesForWellConnectionPressures(const Simulator& simulator,
                                                          const WellState& well_state,
-                                                         std::vector<double>& b_perf,
-                                                         std::vector<double>& rsmax_perf,
-                                                         std::vector<double>& rvmax_perf,
-                                                         std::vector<double>& rvwmax_perf,
-                                                         std::vector<double>& rswmax_perf,
-                                                         std::vector<double>& surf_dens_perf) const;
+                                                         WellConnectionProps& props) const;
 
-        void computeWellConnectionDensitesPressures(const Simulator& ebosSimulator,
+        void computeWellConnectionDensitesPressures(const Simulator& simulator,
                                                     const WellState& well_state,
-                                                    const std::vector<double>& b_perf,
-                                                    const std::vector<double>& rsmax_perf,
-                                                    const std::vector<double>& rvmax_perf,
-                                                    const std::vector<double>& rvwmax_perf,
-                                                    const std::vector<double>& rswmax_perf,
-                                                    const std::vector<double>& surf_dens_perf,
+                                                    const WellConnectionProps& props,
                                                     DeferredLogger& deferred_logger);
 
-        void computeWellConnectionPressures(const Simulator& ebosSimulator,
+        void computeWellConnectionPressures(const Simulator& simulator,
                                             const WellState& well_state,
                                             DeferredLogger& deferred_logger);
 
-        void computePerfRateEval(const IntensiveQuantities& intQuants,
-                                 const std::vector<EvalWell>& mob,
-                                 const EvalWell& bhp,
-                                 const double Tw,
-                                 const int perf,
-                                 const bool allow_cf,
-                                 std::vector<EvalWell>& cq_s,
-                                 double& perf_dis_gas_rate,
-                                 double& perf_dis_gas_rate_in_water,
-                                 double& perf_vap_oil_rate,
-                                 double& perf_vap_wat_rate,
-                                 DeferredLogger& deferred_logger) const;
-
-        void computePerfRateScalar(const IntensiveQuantities& intQuants,
-                                   const std::vector<Scalar>& mob,
-                                   const Scalar& bhp,
-                                   const double Tw,
-                                   const int perf,
-                                   const bool allow_cf,
-                                   std::vector<Scalar>& cq_s,
-                                   DeferredLogger& deferred_logger) const;
+        template<class Value>
+        void computePerfRate(const IntensiveQuantities& intQuants,
+                             const std::vector<Value>& mob,
+                             const Value& bhp,
+                             const std::vector<Scalar>& Tw,
+                             const int perf,
+                             const bool allow_cf,
+                             std::vector<Value>& cq_s,
+                             PerforationRates& perf_rates,
+                             DeferredLogger& deferred_logger) const;
 
         template<class Value>
         void computePerfRate(const std::vector<Value>& mob,
@@ -318,48 +306,39 @@ namespace Opm
                              const Value& rvw,
                              const Value& rsw,
                              std::vector<Value>& b_perfcells_dense,
-                             const double Tw,
+                             const std::vector<Scalar>& Tw,
                              const int perf,
                              const bool allow_cf,
                              const Value& skin_pressure,
                              const std::vector<Value>& cmix_s,
                              std::vector<Value>& cq_s,
-                             double& perf_dis_gas_rate,
-                             double& perf_dis_gas_rate_in_water,
-                             double& perf_vap_oil_rate,
-                             double& perf_vap_wat_rate,
+                             PerforationRates& perf_rates,
                              DeferredLogger& deferred_logger) const;
 
-        void computeWellRatesWithBhpIterations(const Simulator& ebosSimulator,
+        void computeWellRatesWithBhpIterations(const Simulator& simulator,
                                               const double& bhp,
                                               std::vector<double>& well_flux,
-                                              DeferredLogger& deferred_logger) const;
+                                              DeferredLogger& deferred_logger) const override;
 
         std::vector<double> computeWellPotentialWithTHP(
-            const Simulator& ebosSimulator,
+            const Simulator& simulator,
             DeferredLogger& deferred_logger,
             const WellState &well_state) const;
 
-        bool updateWellStateWithTHPTargetProd(const Simulator& ebos_simulator,
-                                              WellState& well_state,
-                                              DeferredLogger& deferred_logger) const override;
+        bool computeWellPotentialsImplicit(const Simulator& simulator,
+                                           std::vector<double>& well_potentials,
+                                           DeferredLogger& deferred_logger) const;               
 
         virtual double getRefDensity() const override;
 
         // get the mobility for specific perforation
-        void getMobilityEval(const Simulator& ebosSimulator,
-                             const int perf,
-                             std::vector<EvalWell>& mob,
-                             DeferredLogger& deferred_logger) const;
+        template<class Value>
+        void getMobility(const Simulator& simulator,
+                         const int perf,
+                         std::vector<Value>& mob,
+                         DeferredLogger& deferred_logger) const;
 
-        // get the mobility for specific perforation
-        void getMobilityScalar(const Simulator& ebosSimulator,
-                               const int perf,
-                               std::vector<Scalar>& mob,
-                               DeferredLogger& deferred_logger) const;
-
-
-        void updateWaterMobilityWithPolymer(const Simulator& ebos_simulator,
+        void updateWaterMobilityWithPolymer(const Simulator& simulator,
                                             const int perf,
                                             std::vector<EvalWell>& mob_water,
                                             DeferredLogger& deferred_logger) const;
@@ -370,9 +349,10 @@ namespace Opm
 
         void updateWellStateFromPrimaryVariables(const bool stop_or_zero_rate_target,
                                                  WellState& well_state,
+                                                 const SummaryState& summary_state,
                                                  DeferredLogger& deferred_logger) const;
 
-        virtual void assembleWellEqWithoutIteration(const Simulator& ebosSimulator,
+        virtual void assembleWellEqWithoutIteration(const Simulator& simulator,
                                                     const double dt,
                                                     const Well::InjectionControls& inj_controls,
                                                     const Well::ProductionControls& prod_controls,
@@ -380,7 +360,7 @@ namespace Opm
                                                     const GroupState& group_state,
                                                     DeferredLogger& deferred_logger) override;
 
-        void assembleWellEqWithoutIterationImpl(const Simulator& ebosSimulator,
+        void assembleWellEqWithoutIterationImpl(const Simulator& simulator,
                                                 const double dt,
                                                 const Well::InjectionControls& inj_controls,
                                                 const Well::ProductionControls& prod_controls,
@@ -388,7 +368,7 @@ namespace Opm
                                                 const GroupState& group_state,
                                                 DeferredLogger& deferred_logger);
 
-        void calculateSinglePerf(const Simulator& ebosSimulator,
+        void calculateSinglePerf(const Simulator& simulator,
                                  const int perf,
                                  WellState& well_state,
                                  std::vector<RateVector>& connectionRates,
@@ -398,20 +378,25 @@ namespace Opm
                                  DeferredLogger& deferred_logger) const;
 
         // check whether the well is operable under BHP limit with current reservoir condition
-        virtual void checkOperabilityUnderBHPLimit(const WellState& well_state, const Simulator& ebos_simulator, DeferredLogger& deferred_logger) override;
+        void checkOperabilityUnderBHPLimit(const WellState& well_state,
+                                           const Simulator& simulator,
+                                           DeferredLogger& deferred_logger) override;
 
         // check whether the well is operable under THP limit with current reservoir condition
-        virtual void checkOperabilityUnderTHPLimit(const Simulator& ebos_simulator, const WellState& well_state, DeferredLogger& deferred_logger) override;
+        void checkOperabilityUnderTHPLimit(const Simulator& simulator,
+                                           const WellState& well_state,
+                                           DeferredLogger& deferred_logger) override;
 
         // updating the inflow based on the current reservoir condition
-        virtual void updateIPR(const Simulator& ebos_simulator, DeferredLogger& deferred_logger) const override;
+        void updateIPR(const Simulator& simulator,
+                       DeferredLogger& deferred_logger) const override;
 
         // for a well, when all drawdown are in the wrong direction, then this well will not
         // be able to produce/inject .
-        bool allDrawDownWrongDirection(const Simulator& ebos_simulator) const;
+        bool allDrawDownWrongDirection(const Simulator& simulator) const;
 
         // whether the well can produce / inject based on the current well state (bhp)
-        bool canProduceInjectWithCurrentBhp(const Simulator& ebos_simulator,
+        bool canProduceInjectWithCurrentBhp(const Simulator& simulator,
                                             const WellState& well_state,
                                             DeferredLogger& deferred_logger);
 
@@ -420,7 +405,7 @@ namespace Opm
         // we turn on crossflow to avoid singular well equations. It can result in wrong-signed
         // well rates, it can cause problem for THP calculation
         // TODO: looking for better alternative to avoid wrong-signed well rates
-        bool openCrossFlowAvoidSingularity(const Simulator& ebos_simulator) const;
+        bool openCrossFlowAvoidSingularity(const Simulator& simulator) const;
 
         // calculate the skin pressure based on water velocity, throughput and polymer concentration.
         // throughput is used to describe the formation damage during water/polymer injection.
@@ -442,12 +427,12 @@ namespace Opm
                             DeferredLogger& deferred_logger) const;
 
         // modify the water rate for polymer injectivity study
-        void handleInjectivityRate(const Simulator& ebosSimulator,
+        void handleInjectivityRate(const Simulator& simulator,
                                    const int perf,
                                    std::vector<EvalWell>& cq_s) const;
 
         // handle the extra equations for polymer injectivity study
-        void handleInjectivityEquations(const Simulator& ebosSimulator,
+        void handleInjectivityEquations(const Simulator& simulator,
                                         const WellState& well_state,
                                         const int perf,
                                         const EvalWell& water_flux_s,
@@ -469,14 +454,67 @@ namespace Opm
 
 
         std::optional<double> computeBhpAtThpLimitProd(const WellState& well_state,
-                                                       const Simulator& ebos_simulator,
+                                                       const Simulator& simulator,
                                                        const SummaryState& summary_state,
                                                        DeferredLogger& deferred_logger) const;
 
-        std::optional<double> computeBhpAtThpLimitInj(const Simulator& ebos_simulator,
+        std::optional<double> computeBhpAtThpLimitInj(const Simulator& simulator,
                                                       const SummaryState& summary_state,
                                                       DeferredLogger& deferred_logger) const;
 
+    private:
+        Eval connectionRateEnergy(const double maxOilSaturation,
+                                  const std::vector<EvalWell>& cq_s,
+                                  const IntensiveQuantities& intQuants,
+                                  DeferredLogger& deferred_logger) const;
+
+        template<class Value>
+        void gasOilPerfRateInj(const std::vector<Value>& cq_s,
+                               PerforationRates& perf_rates,
+                               const Value& rv,
+                               const Value& rs,
+                               const Value& pressure,
+                               const Value& rvw,
+                               DeferredLogger& deferred_logger) const;
+
+        template<class Value>
+        void gasOilPerfRateProd(std::vector<Value>& cq_s,
+                                PerforationRates& perf_rates,
+                                const Value& rv,
+                                const Value& rs,
+                                const Value& rvw) const;
+
+        template<class Value>
+        void gasWaterPerfRateProd(std::vector<Value>& cq_s,
+                                  PerforationRates& perf_rates,
+                                  const Value& rvw,
+                                  const Value& rsw) const;
+
+        template<class Value>
+        void gasWaterPerfRateInj(const std::vector<Value>& cq_s,
+                                 PerforationRates& perf_rates,
+                                 const Value& rvw,
+                                 const Value& rsw,
+                                 const Value& pressure,
+                                 DeferredLogger& deferred_logger) const;
+
+        template<class Value>
+        void disOilVapWatVolumeRatio(Value& volumeRatio,
+                                     const Value& rvw,
+                                     const Value& rsw,
+                                     const Value& pressure,
+                                     const std::vector<Value>& cmix_s,
+                                     const std::vector<Value>& b_perfcells_dense,
+                                     DeferredLogger& deferred_logger) const;
+
+        template<class Value>
+        void gasOilVolumeRatio(Value& volumeRatio,
+                               const Value& rv,
+                               const Value& rs,
+                               const Value& pressure,
+                               const std::vector<Value>& cmix_s,
+                               const std::vector<Value>& b_perfcells_dense,
+                               DeferredLogger& deferred_logger) const;
     };
 
 }
