@@ -30,28 +30,30 @@
 #include <opm/models/blackoil/blackoilonephaseindices.hh>
 #include <opm/models/blackoil/blackoiltwophaseindices.hh>
 
-#include <opm/simulators/utils/DeferredLogger.hpp>
+#include <opm/simulators/utils/DeferredLoggingErrorHelpers.hpp>
 #include <opm/simulators/wells/ParallelWellInfo.hpp>
 #include <opm/simulators/wells/WellInterfaceIndices.hpp>
 #include <opm/simulators/wells/WellState.hpp>
 
+#include <cassert>
+#include <limits>
 #include <numeric>
 #include <sstream>
 
 namespace Opm
 {
 
-template<class FluidSystem, class Indices, class Scalar>
-StandardWellConnections<FluidSystem,Indices,Scalar>::
-StandardWellConnections(const WellInterfaceIndices<FluidSystem,Indices,Scalar>& well)
+template<class FluidSystem, class Indices>
+StandardWellConnections<FluidSystem,Indices>::
+StandardWellConnections(const WellInterfaceIndices<FluidSystem,Indices>& well)
     : well_(well)
     , perf_densities_(well.numPerfs())
     , perf_pressure_diffs_(well.numPerfs())
 {
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void StandardWellConnections<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void StandardWellConnections<FluidSystem,Indices>::
 computePressureDelta()
 {
     // Algorithm:
@@ -86,15 +88,10 @@ computePressureDelta()
     well_.parallelWellInfo().partialSumPerfValues(beg, end);
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void StandardWellConnections<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void StandardWellConnections<FluidSystem,Indices>::
 computeDensities(const std::vector<Scalar>& perfComponentRates,
-                 const std::vector<Scalar>& b_perf,
-                 const std::vector<Scalar>& rsmax_perf,
-                 const std::vector<Scalar>& rvmax_perf,
-                 const std::vector<Scalar>& rvwmax_perf,
-                 const std::vector<Scalar>& rswmax_perf,
-                 const std::vector<Scalar>& surf_dens_perf,
+                 const Properties& props,
                  DeferredLogger& deferred_logger)
 {
     // Verify that we have consistent input.
@@ -208,11 +205,11 @@ computeDensities(const std::vector<Scalar>& perfComponentRates,
             const unsigned oilpos = Indices::canonicalToActiveComponentIndex(FluidSystem::oilCompIdx);
             Scalar rs = 0.0;
             Scalar rv = 0.0;
-            if (!rsmax_perf.empty() && mix[oilpos] > 1e-12) {
-                rs = std::min(mix[gaspos]/mix[oilpos], rsmax_perf[perf]);
+            if (!props.rsmax_perf.empty() && mix[oilpos] > 1e-12) {
+                rs = std::min(mix[gaspos] / mix[oilpos], props.rsmax_perf[perf]);
             }
-            if (!rvmax_perf.empty() && mix[gaspos] > 1e-12) {
-                rv = std::min(mix[oilpos]/mix[gaspos], rvmax_perf[perf]);
+            if (!props.rvmax_perf.empty() && mix[gaspos] > 1e-12) {
+                rv = std::min(mix[oilpos] / mix[gaspos], props.rvmax_perf[perf]);
             }
             const Scalar d = 1.0 - rs*rv;
             if (d <= 0.0) {
@@ -246,12 +243,12 @@ computeDensities(const std::vector<Scalar>& perfComponentRates,
             const unsigned waterpos = Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx);
             const unsigned gaspos = Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx);
             Scalar rvw = 0.0;
-            if (!rvwmax_perf.empty() && mix[gaspos] > 1e-12) {
-                rvw = std::min(mix[waterpos]/mix[gaspos], rvwmax_perf[perf]);
+            if (!props.rvwmax_perf.empty() && mix[gaspos] > 1e-12) {
+                rvw = std::min(mix[waterpos] / mix[gaspos], props.rvwmax_perf[perf]);
             }
             Scalar rsw = 0.0;
-            if (!rswmax_perf.empty() && mix[waterpos] > 1e-12) {
-                rsw = std::min(mix[gaspos]/mix[waterpos], rswmax_perf[perf]);
+            if (!props.rswmax_perf.empty() && mix[waterpos] > 1e-12) {
+                rsw = std::min(mix[gaspos] / mix[waterpos], props.rswmax_perf[perf]);
             }
             const Scalar d = 1.0 - rsw*rvw;
             if (d <= 0.0) {
@@ -277,10 +274,10 @@ computeDensities(const std::vector<Scalar>& perfComponentRates,
 
         Scalar volrat = 0.0;
         for (int component = 0; component < num_comp; ++component) {
-            volrat += x[component] / b_perf[perf*num_comp+ component];
+            volrat += x[component] / props.b_perf[perf * num_comp + component];
         }
         for (int component = 0; component < num_comp; ++component) {
-            surf_dens[component] = surf_dens_perf[perf*num_comp+ component];
+            surf_dens[component] = props.surf_dens_perf[perf * num_comp + component];
         }
 
         // Compute segment density.
@@ -288,25 +285,20 @@ computeDensities(const std::vector<Scalar>& perfComponentRates,
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void StandardWellConnections<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void StandardWellConnections<FluidSystem,Indices>::
 computePropertiesForPressures(const WellState& well_state,
                               const std::function<Scalar(int,int)>& getTemperature,
                               const std::function<Scalar(int)>& getSaltConcentration,
                               const std::function<int(int)>& pvtRegionIdx,
                               const std::function<Scalar(int)>& solventInverseFormationVolumeFactor,
                               const std::function<Scalar(int)>& solventRefDensity,
-                              std::vector<Scalar>& b_perf,
-                              std::vector<Scalar>& rsmax_perf,
-                              std::vector<Scalar>& rvmax_perf,
-                              std::vector<Scalar>& rvwmax_perf,
-                              std::vector<Scalar>& rswmax_perf,
-                              std::vector<Scalar>& surf_dens_perf) const
+                              Properties& props) const
 {
     const int nperf = well_.numPerfs();
     const PhaseUsage& pu = well_.phaseUsage();
-    b_perf.resize(nperf * well_.numComponents());
-    surf_dens_perf.resize(nperf * well_.numComponents());
+    props.b_perf.resize(nperf * well_.numComponents());
+    props.surf_dens_perf.resize(nperf * well_.numComponents());
     const auto& ws = well_state.well(well_.indexOfWell());
 
     static constexpr int Water = BlackoilPhases::Aqua;
@@ -319,13 +311,13 @@ computePropertiesForPressures(const WellState& well_state,
 
     //rs and rv are only used if both oil and gas is present
     if (oilPresent && gasPresent) {
-        rsmax_perf.resize(nperf);
-        rvmax_perf.resize(nperf);
+        props.rsmax_perf.resize(nperf);
+        props.rvmax_perf.resize(nperf);
     }
       //rvw is only used if both water and gas is present
     if (waterPresent && gasPresent) {
-        rvwmax_perf.resize(nperf);
-        rswmax_perf.resize(nperf);
+        props.rvwmax_perf.resize(nperf);
+        props.rswmax_perf.resize(nperf);
     }
 
     // Compute the average pressure in each well block
@@ -349,109 +341,64 @@ computePropertiesForPressures(const WellState& well_state,
                 // TODO support mutual solubility in water and oil
                 assert(!FluidSystem::phaseIsActive(FluidSystem::oilPhaseIdx));
                 const double waterrate = std::abs(ws.surface_rates[pu.phase_pos[Water]]);
-                rswmax_perf[perf] = FluidSystem::waterPvt().saturatedGasDissolutionFactor(region_idx, temperature, p_avg, saltConcentration);
+                props.rswmax_perf[perf] = FluidSystem::waterPvt().saturatedGasDissolutionFactor(region_idx, temperature, p_avg, saltConcentration);
                 if (waterrate > 0) {
                     const double gasrate = std::abs(ws.surface_rates[pu.phase_pos[Gas]]) - (Indices::enableSolvent ? ws.sum_solvent_rates() : 0.0);
                     if (gasrate > 0) {
-                        rsw = waterrate / gasrate;
+                        rsw = gasrate / waterrate;
                     }
-                    rsw = std::min(rsw, rswmax_perf[perf]);
+                    rsw = std::min(rsw, props.rswmax_perf[perf]);
                 }
             }
-            b_perf[ waterCompIdx + perf * well_.numComponents()] = FluidSystem::waterPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, rsw, saltConcentration);
+            props.b_perf[waterCompIdx + perf * well_.numComponents()] = FluidSystem::waterPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, rsw, saltConcentration);
         }
 
         if (gasPresent) {
             const unsigned gasCompIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx);
             const int gaspos = gasCompIdx + perf * well_.numComponents();
-
-            if (oilPresent && waterPresent) {
+            double rvw = 0.0;
+            double rv = 0.0;
+            if (oilPresent) {
                 const double oilrate = std::abs(ws.surface_rates[pu.phase_pos[Oil]]); //in order to handle negative rates in producers
-                const double waterrate = std::abs(ws.surface_rates[pu.phase_pos[Water]]); //in order to handle negative rates in producers
-                rvmax_perf[perf] = FluidSystem::gasPvt().saturatedOilVaporizationFactor(region_idx, temperature, p_avg);
-                rvwmax_perf[perf] = FluidSystem::gasPvt().saturatedWaterVaporizationFactor(region_idx, temperature, p_avg);
-                double rv = 0.0;
-                double rvw = 0.0;
+                props.rvmax_perf[perf] = FluidSystem::gasPvt().saturatedOilVaporizationFactor(region_idx, temperature, p_avg);
                 if (oilrate > 0) {
                     const double gasrate = std::abs(ws.surface_rates[pu.phase_pos[Gas]]) - (Indices::enableSolvent ? ws.sum_solvent_rates() : 0.0);
                     if (gasrate > 0) {
                         rv = oilrate / gasrate;
                     }
-                    rv = std::min(rv, rvmax_perf[perf]);
+                    rv = std::min(rv, props.rvmax_perf[perf]);
                 }
-                if (waterrate > 0) {
-                    const double gasrate = std::abs(ws.surface_rates[pu.phase_pos[Gas]]) - (Indices::enableSolvent ? ws.sum_solvent_rates() : 0.0);
-                    if (gasrate > 0) {
-                        rvw = waterrate / gasrate;
-                    }
-                    rvw = std::min(rvw, rvwmax_perf[perf]);
-                }
-                if (rv > 0.0 || rvw > 0.0){
-                    b_perf[gaspos] = FluidSystem::gasPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, rv, rvw);
-                }
-                else {
-                    b_perf[gaspos] = FluidSystem::gasPvt().saturatedInverseFormationVolumeFactor(region_idx, temperature, p_avg);
-                }
-            } else if (oilPresent) {
-                //no water
-                const double oilrate = std::abs(ws.surface_rates[pu.phase_pos[Oil]]); //in order to handle negative rates in producers
-                rvmax_perf[perf] = FluidSystem::gasPvt().saturatedOilVaporizationFactor(region_idx, temperature, p_avg);
-                if (oilrate > 0) {
-                    const double gasrate = std::abs(ws.surface_rates[pu.phase_pos[Gas]]) - (Indices::enableSolvent ? ws.sum_solvent_rates() : 0.0);
-                    double rv = 0.0;
-                    if (gasrate > 0) {
-                        rv = oilrate / gasrate;
-                    }
-                    rv = std::min(rv, rvmax_perf[perf]);
-
-                    b_perf[gaspos] = FluidSystem::gasPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, rv, 0.0 /*Rvw*/);
-                }
-                else {
-                    b_perf[gaspos] = FluidSystem::gasPvt().saturatedInverseFormationVolumeFactor(region_idx, temperature, p_avg);
-                }
-            } else if (waterPresent) {
-                //no oil
-                const double waterrate = std::abs(ws.surface_rates[pu.phase_pos[Water]]); //in order to handle negative rates in producers
-                rvwmax_perf[perf] = FluidSystem::gasPvt().saturatedWaterVaporizationFactor(region_idx, temperature, p_avg);
-                if (waterrate > 0) {
-                    const double gasrate = std::abs(ws.surface_rates[pu.phase_pos[Gas]]) - (Indices::enableSolvent ? ws.sum_solvent_rates() : 0.0);
-                    double rvw = 0.0;
-                    if (gasrate > 0) {
-                        rvw = waterrate / gasrate;
-                    }
-                    rvw = std::min(rvw, rvwmax_perf[perf]);
-
-                    b_perf[gaspos] = FluidSystem::gasPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, 0.0 /*Rv*/, rvw);
-                }
-                else {
-                    b_perf[gaspos] = FluidSystem::gasPvt().saturatedInverseFormationVolumeFactor(region_idx, temperature, p_avg);
-                }
-
-            } else {
-                b_perf[gaspos] = FluidSystem::gasPvt().saturatedInverseFormationVolumeFactor(region_idx, temperature, p_avg);
             }
+            if (waterPresent) {
+                const double waterrate = std::abs(ws.surface_rates[pu.phase_pos[Water]]); //in order to handle negative rates in producers
+                props.rvwmax_perf[perf] = FluidSystem::gasPvt().saturatedWaterVaporizationFactor(region_idx, temperature, p_avg);
+                if (waterrate > 0) {
+                    const double gasrate = std::abs(ws.surface_rates[pu.phase_pos[Gas]]) - (Indices::enableSolvent ? ws.sum_solvent_rates() : 0.0);
+                    if (gasrate > 0) {
+                        rvw = waterrate / gasrate;
+                    }
+                    rvw = std::min(rvw, props.rvwmax_perf[perf]);
+                }
+            }
+            props.b_perf[gaspos] = FluidSystem::gasPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, rv, rvw);
         }
 
         if (oilPresent) {
             const unsigned oilCompIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::oilCompIdx);
             const int oilpos = oilCompIdx + perf * well_.numComponents();
+            double rs = 0.0;
             if (gasPresent) {
-                rsmax_perf[perf] = FluidSystem::oilPvt().saturatedGasDissolutionFactor(region_idx, temperature, p_avg);
+                props.rsmax_perf[perf] = FluidSystem::oilPvt().saturatedGasDissolutionFactor(region_idx, temperature, p_avg);
                 const double gasrate = std::abs(ws.surface_rates[pu.phase_pos[Gas]]) - (Indices::enableSolvent ? ws.sum_solvent_rates() : 0.0);
                 if (gasrate > 0) {
                     const double oilrate = std::abs(ws.surface_rates[pu.phase_pos[Oil]]);
-                    double rs = 0.0;
                     if (oilrate > 0) {
                         rs = gasrate / oilrate;
                     }
-                    rs = std::min(rs, rsmax_perf[perf]);
-                    b_perf[oilpos] = FluidSystem::oilPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, rs);
-                } else {
-                    b_perf[oilpos] = FluidSystem::oilPvt().saturatedInverseFormationVolumeFactor(region_idx, temperature, p_avg);
+                    rs = std::min(rs, props.rsmax_perf[perf]);
                 }
-            } else {
-                b_perf[oilpos] = FluidSystem::oilPvt().saturatedInverseFormationVolumeFactor(region_idx, temperature, p_avg);
             }
+            props.b_perf[oilpos] = FluidSystem::oilPvt().inverseFormationVolumeFactor(region_idx, temperature, p_avg, rs);
         }
 
         // Surface density.
@@ -461,43 +408,38 @@ computePropertiesForPressures(const WellState& well_state,
             }
 
             const unsigned compIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::solventComponentIndex(phaseIdx));
-            surf_dens_perf[well_.numComponents() * perf  + compIdx] = FluidSystem::referenceDensity( phaseIdx, region_idx );
+            props.surf_dens_perf[well_.numComponents() * perf  + compIdx] = FluidSystem::referenceDensity( phaseIdx, region_idx );
         }
 
         // We use cell values for solvent injector
         if constexpr (Indices::enableSolvent) {
-            b_perf[well_.numComponents() * perf + Indices::contiSolventEqIdx] = solventInverseFormationVolumeFactor(cell_idx);
-            surf_dens_perf[well_.numComponents() * perf + Indices::contiSolventEqIdx] = solventRefDensity(cell_idx);
+            props.b_perf[well_.numComponents() * perf + Indices::contiSolventEqIdx] = solventInverseFormationVolumeFactor(cell_idx);
+            props.surf_dens_perf[well_.numComponents() * perf + Indices::contiSolventEqIdx] = solventRefDensity(cell_idx);
         }
     }
 }
 
-template<class FluidSystem, class Indices, class Scalar>
-void StandardWellConnections<FluidSystem,Indices,Scalar>::
+template<class FluidSystem, class Indices>
+void StandardWellConnections<FluidSystem,Indices>::
 computeProperties(const WellState& well_state,
                   const std::function<Scalar(int,int)>& invB,
                   const std::function<Scalar(int,int)>& mobility,
                   const std::function<Scalar(int)>& solventInverseFormationVolumeFactor,
                   const std::function<Scalar(int)>& solventMobility,
-                  const std::vector<Scalar>& b_perf,
-                  const std::vector<Scalar>& rsmax_perf,
-                  const std::vector<Scalar>& rvmax_perf,
-                  const std::vector<Scalar>& rvwmax_perf,
-                  const std::vector<Scalar>& rswmax_perf,
-                  const std::vector<Scalar>& surf_dens_perf,
+                  const Properties& props,
                   DeferredLogger& deferred_logger)
 {
     // Compute densities
     const int nperf = well_.numPerfs();
     const int np = well_.numPhases();
-    std::vector<double> perfRates(b_perf.size(),0.0);
+    std::vector<double> perfRates(props.b_perf.size(),0.0);
     const auto& ws = well_state.well(well_.indexOfWell());
     const auto& perf_data = ws.perf_data;
     const auto& perf_rates_state = perf_data.phase_rates;
 
     for (int perf = 0; perf < nperf; ++perf) {
         for (int comp = 0; comp < np; ++comp) {
-            perfRates[perf * well_.numComponents() + comp] =  perf_rates_state[perf * np + well_.ebosCompIdxToFlowCompIdx(comp)];
+            perfRates[perf * well_.numComponents() + comp] =  perf_rates_state[perf * np + well_.modelCompIdxToFlowCompIdx(comp)];
         }
     }
 
@@ -528,33 +470,204 @@ computeProperties(const WellState& well_state,
         {
             total_tw = comm.sum(total_tw);
         }
+        // for producers where all perforations have zero rates we
+        // approximate the perforation mixture ration using the (invB * mobility) ratio,
+        // and weight the perforation rates using the well transmissibility.
         for (int perf = 0; perf < nperf; ++perf) {
             const int cell_idx = well_.cells()[perf];
             const double well_tw_fraction = well_.wellIndex()[perf] / total_tw;
             double total_mobility = 0.0;
+            double total_invB = 0.;
             for (int p = 0; p < np; ++p) {
-                int ebosPhaseIdx = well_.flowPhaseToEbosPhaseIdx(p);
-                total_mobility += invB(cell_idx, ebosPhaseIdx) * mobility(cell_idx, ebosPhaseIdx);
+                int modelPhaseIdx = well_.flowPhaseToModelPhaseIdx(p);
+                total_mobility += invB(cell_idx, modelPhaseIdx) * mobility(cell_idx, modelPhaseIdx);
+                total_invB += invB(cell_idx, modelPhaseIdx);
             }
             if constexpr (Indices::enableSolvent) {
                 total_mobility += solventInverseFormationVolumeFactor(cell_idx) * solventMobility(cell_idx);
+                total_invB += solventInverseFormationVolumeFactor(cell_idx);
             }
+            const bool non_zero_total_mobility = total_mobility > std::numeric_limits<double>::min();
+            assert(total_invB > std::numeric_limits<double>::min());
+            // for the perforation having zero mobility for all the phases, we use a small value to generate a small
+            // perforation rates for those perforations, at the same time, we can use the rates to recover the mixing
+            // ratios for those perforations.
+            constexpr double small_value = 1.e-10;
             for (int p = 0; p < np; ++p) {
-                int ebosPhaseIdx = well_.flowPhaseToEbosPhaseIdx(p);
-                perfRates[perf * well_.numComponents() + p] = well_tw_fraction * mobility(cell_idx, ebosPhaseIdx) / total_mobility;
+                const int modelPhaseIdx = well_.flowPhaseToModelPhaseIdx(p);
+                const auto mob_ratio = non_zero_total_mobility
+                                       ? mobility(cell_idx, modelPhaseIdx) / total_mobility
+                                       : small_value / total_invB;
+                perfRates[perf * well_.numComponents() + p] = well_tw_fraction * invB(cell_idx, modelPhaseIdx) * mob_ratio;
             }
             if constexpr (Indices::enableSolvent) {
-                perfRates[perf * well_.numComponents() + Indices::contiSolventEqIdx] = well_tw_fraction * solventInverseFormationVolumeFactor(cell_idx) / total_mobility;
+                const auto mob_ratio = non_zero_total_mobility
+                                        ? solventMobility(cell_idx) / total_mobility
+                                        : small_value / total_invB;
+                perfRates[perf * well_.numComponents() + Indices::contiSolventEqIdx] =
+                        well_tw_fraction * solventInverseFormationVolumeFactor(cell_idx) * mob_ratio;
             }
         }
     }
 
-    this->computeDensities(perfRates, b_perf, rsmax_perf, rvmax_perf, rvwmax_perf, rswmax_perf, surf_dens_perf, deferred_logger);
+    this->computeDensities(perfRates, props, deferred_logger);
     this->computePressureDelta();
 }
 
+template<class FluidSystem, class Indices>
+typename StandardWellConnections<FluidSystem,Indices>::Eval
+StandardWellConnections<FluidSystem,Indices>::
+connectionRateBrine(double& rate,
+                    const double vap_wat_rate,
+                    const std::vector<EvalWell>& cq_s,
+                    const std::variant<Scalar,EvalWell>& saltConcentration) const
+{
+    // TODO: the application of well efficiency factor has not been tested with an example yet
+    const unsigned waterCompIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx);
+    // Correction salt rate; evaporated water does not contain salt
+    EvalWell cq_s_sm = cq_s[waterCompIdx] - vap_wat_rate;
+    if (well_.isInjector()) {
+        cq_s_sm *= std::get<Scalar>(saltConcentration);
+    } else {
+        cq_s_sm *= std::get<EvalWell>(saltConcentration);
+    }
+
+    // Note. Efficiency factor is handled in the output layer
+    rate = cq_s_sm.value();
+
+    cq_s_sm *= well_.wellEfficiencyFactor();
+    return well_.restrictEval(cq_s_sm);
+}
+
+template<class FluidSystem, class Indices>
+typename StandardWellConnections<FluidSystem,Indices>::Eval
+StandardWellConnections<FluidSystem,Indices>::
+connectionRateFoam(const std::vector<EvalWell>& cq_s,
+                    const std::variant<Scalar,EvalWell>& foamConcentration,
+                    const Phase transportPhase,
+                    DeferredLogger& deferred_logger) const
+{
+    // TODO: the application of well efficiency factor has not been tested with an example yet
+    auto getFoamTransportIdx = [&deferred_logger,transportPhase] {
+        switch (transportPhase) {
+            case Phase::WATER: {
+                return Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx);
+            }
+            case Phase::GAS: {
+                return Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx);
+            }
+            case Phase::SOLVENT: {
+                if constexpr (Indices::enableSolvent)
+                    return static_cast<unsigned>(Indices::contiSolventEqIdx);
+                else
+                    OPM_DEFLOG_THROW(std::runtime_error, "Foam transport phase is SOLVENT but SOLVENT is not activated.", deferred_logger);
+            }
+            default: {
+                OPM_DEFLOG_THROW(std::runtime_error, "Foam transport phase must be GAS/WATER/SOLVENT.", deferred_logger);
+            }
+        }
+    };
+
+    EvalWell cq_s_foam = cq_s[getFoamTransportIdx()] * well_.wellEfficiencyFactor();;
+    if (well_.isInjector()) {
+        cq_s_foam *= std::get<Scalar>(foamConcentration);
+    } else {
+        cq_s_foam *= std::get<EvalWell>(foamConcentration);
+    }
+
+    return well_.restrictEval(cq_s_foam);
+}
+
+template<class FluidSystem, class Indices>
+std::tuple<typename StandardWellConnections<FluidSystem,Indices>::Eval,
+           typename StandardWellConnections<FluidSystem,Indices>::Eval,
+           typename StandardWellConnections<FluidSystem,Indices>::Eval>
+StandardWellConnections<FluidSystem,Indices>::
+connectionRatesMICP(const std::vector<EvalWell>& cq_s,
+                    const std::variant<Scalar,EvalWell>& microbialConcentration,
+                    const std::variant<Scalar,EvalWell>& oxygenConcentration,
+                    const std::variant<Scalar,EvalWell>& ureaConcentration) const
+{
+    const unsigned waterCompIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx);
+    EvalWell cq_s_microbe = cq_s[waterCompIdx];
+    if (well_.isInjector()) {
+        cq_s_microbe *= std::get<Scalar>(microbialConcentration);
+    } else {
+        cq_s_microbe *= std::get<EvalWell>(microbialConcentration);
+    }
+
+    EvalWell cq_s_oxygen = cq_s[waterCompIdx];
+    if (well_.isInjector()) {
+        cq_s_oxygen *= std::get<Scalar>(oxygenConcentration);
+    } else {
+        cq_s_oxygen *= std::get<EvalWell>(oxygenConcentration);
+    }
+
+    EvalWell cq_s_urea = cq_s[waterCompIdx];
+    if (well_.isInjector()) {
+        cq_s_urea *= std::get<Scalar>(ureaConcentration);
+    } else {
+        cq_s_urea *= std::get<EvalWell>(ureaConcentration);
+    }
+
+    return {well_.restrictEval(cq_s_microbe),
+            well_.restrictEval(cq_s_oxygen),
+            well_.restrictEval(cq_s_urea)};
+}
+
+template<class FluidSystem, class Indices>
+std::tuple<typename StandardWellConnections<FluidSystem,Indices>::Eval,
+           typename StandardWellConnections<FluidSystem,Indices>::EvalWell>
+StandardWellConnections<FluidSystem,Indices>::
+connectionRatePolymer(double& rate,
+                    const std::vector<EvalWell>& cq_s,
+                    const std::variant<Scalar,EvalWell>& polymerConcentration) const
+{
+    // TODO: the application of well efficiency factor has not been tested with an example yet
+    const unsigned waterCompIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::waterCompIdx);
+    EvalWell cq_s_poly = cq_s[waterCompIdx];
+    if (well_.isInjector()) {
+        cq_s_poly *= std::get<Scalar>(polymerConcentration);
+    } else {
+        cq_s_poly *= std::get<EvalWell>(polymerConcentration);
+    }
+    // Note. Efficiency factor is handled in the output layer
+    rate = cq_s_poly.value();
+
+    cq_s_poly *= well_.wellEfficiencyFactor();
+
+    return {well_.restrictEval(cq_s_poly), cq_s_poly};
+}
+
+template<class FluidSystem, class Indices>
+std::tuple<typename StandardWellConnections<FluidSystem,Indices>::Eval,
+           typename StandardWellConnections<FluidSystem,Indices>::EvalWell>
+StandardWellConnections<FluidSystem,Indices>::
+connectionRatezFraction(double& rate,
+                        const double dis_gas_rate,
+                        const std::vector<EvalWell>& cq_s,
+                        const std::variant<Scalar, std::array<EvalWell,2>>& solventConcentration) const
+{
+    // TODO: the application of well efficiency factor has not been tested with an example yet
+    const unsigned gasCompIdx = Indices::canonicalToActiveComponentIndex(FluidSystem::gasCompIdx);
+    EvalWell cq_s_zfrac_effective = cq_s[gasCompIdx];
+    if (well_.isInjector()) {
+        cq_s_zfrac_effective *= std::get<Scalar>(solventConcentration);
+    } else if (cq_s_zfrac_effective.value() != 0.0) {
+        const double dis_gas_frac = dis_gas_rate / cq_s_zfrac_effective.value();
+        const auto& vol = std::get<std::array<EvalWell,2>>(solventConcentration);
+        cq_s_zfrac_effective *= dis_gas_frac * vol[0] + (1.0 - dis_gas_frac) * vol[1];
+    }
+
+    rate = cq_s_zfrac_effective.value();
+
+    cq_s_zfrac_effective *= well_.wellEfficiencyFactor();
+    return {well_.restrictEval(cq_s_zfrac_effective), cq_s_zfrac_effective};
+}
+
 #define INSTANCE(...) \
-template class StandardWellConnections<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>,__VA_ARGS__,double>;
+template class StandardWellConnections<BlackOilFluidSystem<double,BlackOilDefaultIndexTraits>, \
+                                       __VA_ARGS__>;
 
 // One phase
 INSTANCE(BlackOilOnePhaseIndices<0u,0u,0u,0u,false,false,0u,1u,0u>)
@@ -572,6 +685,9 @@ INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,0u,0u>)
 INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,0u,false,true,0u,2u,0u>)
 INSTANCE(BlackOilTwoPhaseIndices<0u,0u,2u,0u,false,false,0u,2u,0u>)
 INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,false,0u,0u,0u>)
+INSTANCE(BlackOilTwoPhaseIndices<0u,0u,0u,1u,false,true,0u,0u,0u>)
+INSTANCE(BlackOilTwoPhaseIndices<1u,0u,0u,0u,false,false,0u,0u,0u>)
+
 // Blackoil
 INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,false,0u,0u>)
 INSTANCE(BlackOilIndices<1u,0u,0u,0u,false,false,0u,0u>)
@@ -582,5 +698,6 @@ INSTANCE(BlackOilIndices<0u,0u,0u,0u,true,false,0u,0u>)
 INSTANCE(BlackOilIndices<0u,0u,0u,0u,false,true,0u,0u>)
 INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,true,0u,0u>)
 INSTANCE(BlackOilIndices<0u,0u,0u,1u,false,false,1u,0u>)
+INSTANCE(BlackOilIndices<1u,0u,0u,0u,true,false,0u,0u>)
 
 }
