@@ -27,7 +27,7 @@
 
 #include <opm/simulators/wells/VFPHelpers.hpp>
 
-
+#include <cstddef>
 
 namespace Opm {
 
@@ -125,7 +125,7 @@ bhpwithflo(const std::vector<double>& flos,
     const auto alq_i = detail::findInterpData( alq, table.getALQAxis()); //assume constant
 
     std::vector<double> bhps(flos.size(), 0.);
-    for (size_t i = 0; i < flos.size(); ++i) {
+    for (std::size_t i = 0; i < flos.size(); ++i) {
         // Value of FLO is negative in OPM for producers, but positive in VFP table
         const auto flo_i = detail::findInterpData(-flos[i], table.getFloAxis());
         const detail::VFPEvaluation bhp_val = detail::interpolate(table, flo_i, thp_i, wfr_i, gfr_i, alq_i);
@@ -137,6 +137,22 @@ bhpwithflo(const std::vector<double>& flos,
     return bhps;
 }
 
+double
+VFPProdProperties::
+minimumBHP(const int table_id,
+           const double thp,
+           const double wfr,
+           const double gfr,
+           const double alq) const
+{
+    // Get the table
+    const VFPProdTable& table = detail::getTable(m_tables, table_id);
+    const auto retval = detail::getMinimumBHPCoordinate(table, thp, wfr, gfr, alq);
+    // returned pair is (flo, bhp)
+    return retval.second;
+}
+
+
 
 void VFPProdProperties::addTable(const VFPProdTable& new_table) {
     this->m_tables.emplace( new_table.getTableNum(), new_table );
@@ -161,7 +177,7 @@ EvalWell VFPProdProperties::bhp(const int table_id,
     EvalWell flo = detail::getFlo(table, aqua, liquid, vapour);
     EvalWell wfr = detail::getWFR(table, aqua, liquid, vapour);
     EvalWell gfr = detail::getGFR(table, aqua, liquid, vapour);
-    if (use_expvfp) {
+    if (use_expvfp || -flo.value() < table.getFloAxis().front()) {
         wfr = explicit_wfr;
         gfr = explicit_gfr;
     }
@@ -177,7 +193,6 @@ EvalWell VFPProdProperties::bhp(const int table_id,
     detail::VFPEvaluation bhp_val = detail::interpolate(table, flo_i, thp_i, wfr_i, gfr_i, alq_i);
 
     bhp = (bhp_val.dwfr * wfr) + (bhp_val.dgfr * gfr) - (std::max(0.0, bhp_val.dflo) * flo);
-
     bhp.setValue(bhp_val.value);
     return bhp;
 }
@@ -194,6 +209,7 @@ INSTANCE(DenseAd::Evaluation<double, -1, 7u>)
 INSTANCE(DenseAd::Evaluation<double, -1, 8u>)
 INSTANCE(DenseAd::Evaluation<double, -1, 9u>)
 INSTANCE(DenseAd::Evaluation<double, -1, 10u>)
+INSTANCE(DenseAd::Evaluation<double, -1, 11u>)
 INSTANCE(DenseAd::Evaluation<double, 3, 0u>)
 INSTANCE(DenseAd::Evaluation<double, 4, 0u>)
 INSTANCE(DenseAd::Evaluation<double, 5, 0u>)
@@ -201,5 +217,6 @@ INSTANCE(DenseAd::Evaluation<double, 6, 0u>)
 INSTANCE(DenseAd::Evaluation<double, 7, 0u>)
 INSTANCE(DenseAd::Evaluation<double, 8, 0u>)
 INSTANCE(DenseAd::Evaluation<double, 9, 0u>)
+INSTANCE(DenseAd::Evaluation<double, 10, 0u>)
 
 }
