@@ -28,8 +28,8 @@
 
 #include <opm/input/eclipse/Schedule/VFPInjTable.hpp>
 #include <opm/input/eclipse/Schedule/Well/Well.hpp>
-
 #include <opm/input/eclipse/Units/Units.hpp>
+#include <opm/input/eclipse/Schedule/Well/WVFPDP.hpp>
 
 #include <opm/material/densead/Evaluation.hpp>
 
@@ -41,8 +41,12 @@
 
 #include <cassert>
 #include <cmath>
+#include <optional>
+
+#include <fmt/format.h>
 
 static constexpr bool extraBhpAtThpLimitOutput = false;
+static constexpr bool extraThpFromBhpOutput = false;
 
 namespace Opm
 {
@@ -101,7 +105,8 @@ double WellBhpThpCalculator::mostStrictBhpFromBhpLimits(const SummaryState& summ
 double WellBhpThpCalculator::calculateThpFromBhp(const std::vector<double>& rates,
                                                  const double bhp,
                                                  const double rho,
-                                                 const double alq,
+                                                 const std::optional<double>& alq,
+                                                 const double thp_limit,
                                                  DeferredLogger& deferred_logger) const
 {
     assert(int(rates.size()) == 3); // the vfp related only supports three phases now.
@@ -116,24 +121,70 @@ double WellBhpThpCalculator::calculateThpFromBhp(const std::vector<double>& rate
 
     // pick the density in the top layer
     double thp = 0.0;
+    const int table_id = well_.wellEcl().vfp_table_number();
     if (well_.isInjector()) {
-        const int table_id = well_.wellEcl().vfp_table_number();
+        assert(!alq.has_value());
         const double vfp_ref_depth = well_.vfpProperties()->getInj()->getTable(table_id).getDatumDepth();
         const double dp = wellhelpers::computeHydrostaticCorrection(well_.refDepth(), vfp_ref_depth, rho, well_.gravity());
-
-        thp = well_.vfpProperties()->getInj()->thp(table_id, aqua, liquid, vapour, bhp + dp);
+        auto thp_func =
+            [this, table_id, aqua, liquid, vapour, dp](
+                const double bhp_value, const double pressure_loss) {
+                    return this->well_.vfpProperties()->getInj()->thp(
+                           table_id, aqua, liquid, vapour, bhp_value + dp - pressure_loss);
+                };
+        thp = findThpFromBhpIteratively(thp_func, bhp, thp_limit, dp, deferred_logger);
     }
     else if (well_.isProducer()) {
-        const int table_id = well_.wellEcl().vfp_table_number();
         const double vfp_ref_depth = well_.vfpProperties()->getProd()->getTable(table_id).getDatumDepth();
         const double dp = wellhelpers::computeHydrostaticCorrection(well_.refDepth(), vfp_ref_depth, rho, well_.gravity());
-
-        thp = well_.vfpProperties()->getProd()->thp(table_id, aqua, liquid, vapour, bhp + dp, alq);
+        auto thp_func =
+            [this, table_id, aqua, liquid, vapour, dp, &alq]
+                (const double bhp_value, const double pressure_loss) {
+                    return this->well_.vfpProperties()->getProd()->thp(
+                       table_id, aqua, liquid, vapour, bhp_value + dp - pressure_loss, alq.value());
+                };
+        thp = findThpFromBhpIteratively(thp_func, bhp, thp_limit, dp, deferred_logger);
     }
     else {
         OPM_DEFLOG_THROW(std::logic_error, "Expected INJECTOR or PRODUCER well", deferred_logger);
     }
+    return thp;
+}
 
+double
+WellBhpThpCalculator::
+findThpFromBhpIteratively(
+    const std::function<double(const double, const double)>& thp_func,
+    const double bhp,
+    const double thp_limit,
+    const double dp,
+    DeferredLogger& deferred_logger) const
+{
+    auto pressure_loss = getVfpBhpAdjustment(bhp + dp, thp_limit);
+    auto thp = thp_func(bhp, pressure_loss);
+    const double tolerance = 1e-5 * unit::barsa;
+    bool do_iterate = true;
+    int it = 1;
+    int max_iterations = 50;
+    while(do_iterate) {
+        if (it > max_iterations) {
+            break;
+        }
+        double thp_prev = thp;
+        pressure_loss = getVfpBhpAdjustment(bhp + dp - pressure_loss, thp_prev);
+        thp = thp_func(bhp, pressure_loss);
+        auto error = std::fabs(thp-thp_prev);
+        if (extraThpFromBhpOutput) {
+            const std::string msg = fmt::format(
+                "findThpFromBhpIteratively(): iteration {}, thp = {}, bhp = {}, "
+                "pressure_loss = {}, error = {}", it, thp, bhp+dp-pressure_loss, pressure_loss, error);
+            deferred_logger.debug(msg);
+        }
+        if (std::fabs(thp-thp_prev) < tolerance) {
+            break;
+        }
+        it++;
+    }
     return thp;
 }
 
@@ -178,8 +229,16 @@ computeBhpAtThpLimitProd(const std::function<std::vector<double>(const double)>&
         const auto& wfr =  well_.vfpProperties()->getExplicitWFR(controls.vfp_table_number, well_.indexOfWell());
         const auto& gfr = well_.vfpProperties()->getExplicitGFR(controls.vfp_table_number, well_.indexOfWell());
         const bool use_vfpexp = well_.useVfpExplicit();
-        return well_.vfpProperties()->getProd()
-        ->bhp(controls.vfp_table_number, rates[Water], rates[Oil], rates[Gas], thp_limit, alq_value, wfr, gfr, use_vfpexp) - dp;
+        const double bhp = well_.vfpProperties()->getProd()->bhp(controls.vfp_table_number,
+                                                     rates[Water],
+                                                     rates[Oil],
+                                                     rates[Gas],
+                                                     thp_limit,
+                                                     alq_value,
+                                                     wfr,
+                                                     gfr,
+                                                     use_vfpexp);
+        return bhp - dp + getVfpBhpAdjustment(bhp, thp_limit);
     };
 
     // Make the flo() function.
@@ -228,6 +287,7 @@ void WellBhpThpCalculator::updateThp(const double rho,
                                      const std::function<double()>& alq_value,
                                      const std::array<unsigned,3>& active,
                                      WellState& well_state,
+                                     const SummaryState& summary_state,
                                      DeferredLogger& deferred_logger) const
 {
     static constexpr int Gas = BlackoilPhases::Vapour;
@@ -260,8 +320,9 @@ void WellBhpThpCalculator::updateThp(const double rho,
     if (active[Gas]) {
         rates[ Gas ] = ws.surface_rates[pu.phase_pos[ Gas ] ];
     }
-
-    ws.thp = this->calculateThpFromBhp(rates, ws.bhp, rho, alq_value(), deferred_logger);
+    const std::optional<double> alq = this->well_.isProducer() ? std::optional<double>(alq_value()) : std::nullopt;
+    const double thp_limit = well_.getTHPConstraint(summary_state);
+    ws.thp = this->calculateThpFromBhp(rates, ws.bhp, rho, alq, thp_limit, deferred_logger);
 }
 
 template<class EvalWell>
@@ -288,32 +349,74 @@ calculateBhpFromThp(const WellState& well_state,
     const EvalWell aqua = rates[Water];
     const EvalWell liquid = rates[Oil];
     const EvalWell vapour = rates[Gas];
-
-    // pick the reference density
-    // typically the reference in the top layer
+    const double thp_limit = well_.getTHPConstraint(summaryState);
+    double vfp_ref_depth;
+    EvalWell bhp_tab;
     if (well_.isInjector() )
     {
         const auto& controls = well.injectionControls(summaryState);
-        const double vfp_ref_depth = well_.vfpProperties()->getInj()->getTable(controls.vfp_table_number).getDatumDepth();
-        const double dp = wellhelpers::computeHydrostaticCorrection(well_.refDepth(), vfp_ref_depth, rho, well_.gravity());
-        return well_.vfpProperties()->getInj()->bhp(controls.vfp_table_number, aqua, liquid, vapour, well_.getTHPConstraint(summaryState)) - dp;
-     }
-     else if (well_.isProducer()) {
-         const auto& controls = well.productionControls(summaryState);
-         const double vfp_ref_depth = well_.vfpProperties()->getProd()->getTable(controls.vfp_table_number).getDatumDepth();
-         const double dp = wellhelpers::computeHydrostaticCorrection(well_.refDepth(), vfp_ref_depth, rho, well_.gravity());
-         const auto& wfr =  well_.vfpProperties()->getExplicitWFR(controls.vfp_table_number, well_.indexOfWell());
-         const auto& gfr = well_.vfpProperties()->getExplicitGFR(controls.vfp_table_number, well_.indexOfWell());
-         const bool use_vfpexplicit = well_.useVfpExplicit();
-         return well_.vfpProperties()->getProd()->bhp(controls.vfp_table_number,
+        vfp_ref_depth = well_.vfpProperties()->getInj()->getTable(controls.vfp_table_number).getDatumDepth();
+        bhp_tab = well_.vfpProperties()->getInj()->bhp(
+               controls.vfp_table_number, aqua, liquid, vapour, thp_limit);
+    }
+    else if (well_.isProducer()) {
+        const auto& controls = well.productionControls(summaryState);
+        vfp_ref_depth = well_.vfpProperties()->getProd()->getTable(controls.vfp_table_number).getDatumDepth();
+        const auto& wfr =  well_.vfpProperties()->getExplicitWFR(controls.vfp_table_number, well_.indexOfWell());
+        const auto& gfr = well_.vfpProperties()->getExplicitGFR(controls.vfp_table_number, well_.indexOfWell());
+        const bool use_vfpexplicit = well_.useVfpExplicit();
+
+        bhp_tab = well_.vfpProperties()->getProd()->bhp(controls.vfp_table_number,
                                                       aqua, liquid, vapour,
-                                                      well_.getTHPConstraint(summaryState),
+                                                      thp_limit,
                                                       well_.getALQ(well_state),
-                                                      wfr, gfr, use_vfpexplicit) - dp;
-     }
-     else {
-         OPM_DEFLOG_THROW(std::logic_error, "Expected INJECTOR or PRODUCER for well " + well_.name(), deferred_logger);
-     }
+                                                      wfr, gfr, use_vfpexplicit);
+    }
+    else {
+        OPM_DEFLOG_THROW(std::logic_error, "Expected INJECTOR or PRODUCER for well " + well_.name(), deferred_logger);
+    }
+    double bhp_tab_double_value;
+    if constexpr (std::is_same_v<EvalWell, double>) {
+        bhp_tab_double_value = bhp_tab;
+    }
+    else {  // EvalWell and bhp_tab is of type DenseAd::Evaluation<double,...,...>
+        bhp_tab_double_value = bhp_tab.value();
+    }
+    const auto bhp_adjustment = getVfpBhpAdjustment(bhp_tab_double_value, thp_limit);
+    const double dp_hydro = wellhelpers::computeHydrostaticCorrection(
+            well_.refDepth(), vfp_ref_depth, rho, well_.gravity());
+    return bhp_tab - dp_hydro + bhp_adjustment;
+}
+
+double
+WellBhpThpCalculator::
+calculateMinimumBhpFromThp(const WellState& well_state,
+                           const Well& well,
+                           const SummaryState& summaryState,
+                           const double rho) const
+{
+    assert(well_.isProducer()); // only producers can go here for now
+
+    const double thp_limit = well_.getTHPConstraint(summaryState);
+    
+    const auto& controls = well.productionControls(summaryState);
+    const auto& wfr =  well_.vfpProperties()->getExplicitWFR(controls.vfp_table_number, well_.indexOfWell());
+    const auto& gfr = well_.vfpProperties()->getExplicitGFR(controls.vfp_table_number, well_.indexOfWell());
+
+    const double bhp_min = well_.vfpProperties()->getProd()->minimumBHP(controls.vfp_table_number,
+                                                                        thp_limit, wfr, gfr, 
+                                                                        well_.getALQ(well_state));
+ 
+    const double vfp_ref_depth = well_.vfpProperties()->getProd()->getTable(controls.vfp_table_number).getDatumDepth();
+    const auto bhp_adjustment = getVfpBhpAdjustment(bhp_min, thp_limit);
+    const double dp_hydro = wellhelpers::computeHydrostaticCorrection(well_.refDepth(), vfp_ref_depth, 
+                                                                      rho, well_.gravity());
+    return bhp_min - dp_hydro + bhp_adjustment;
+}
+
+double WellBhpThpCalculator::getVfpBhpAdjustment(const double bhp_tab, const double thp_limit) const
+{
+    return well_.wellEcl().getWVFPDP().getPressureLoss(bhp_tab, thp_limit);
 }
 
 template<class ErrorPolicy>
@@ -379,8 +482,9 @@ computeBhpAtThpLimitInjImpl(const std::function<std::vector<double>(const double
                                                                 rho, well_.gravity());
     auto fbhp = [this, &controls, thp_limit, dp](const std::vector<double>& rates) {
         assert(rates.size() == 3);
-        return well_.vfpProperties()->getInj()
-                ->bhp(controls.vfp_table_number, rates[Water], rates[Oil], rates[Gas], thp_limit) - dp;
+        const auto bhp = well_.vfpProperties()->getInj()
+                ->bhp(controls.vfp_table_number, rates[Water], rates[Oil], rates[Gas], thp_limit);
+        return bhp - dp + getVfpBhpAdjustment(bhp, thp_limit);
     };
 
     // Make the flo() function.
@@ -695,6 +799,7 @@ bisectBracket(const std::function<double(const double)>& eq,
                 eq_high = eq(high);
                 abs_high = std::fabs(eq_high);
             }
+            interval = high - low;
             ++bracket_attempts;
         }
 
@@ -736,29 +841,130 @@ bruteForceBracket(const std::function<double(const double)>& eq,
                   double& low, double& high,
                   DeferredLogger& deferred_logger)
 {
-    bool finding_bracket = false;
+    bool bracket_found = false;
     low = range[0];
     high = range[1];
     const int sample_number = 200;
     const double interval = (high - low) / sample_number;
     double eq_low = eq(low);
-    double eq_high;
+    double eq_high = 0.0;
     for (int i = 0; i < sample_number + 1; ++i) {
         high = range[0] + interval * i;
         eq_high = eq(high);
         if (eq_high * eq_low <= 0.) {
-            finding_bracket = true;
+            bracket_found = true;
             break;
         }
         low = high;
         eq_low = eq_high;
     }
-    if (finding_bracket) {
+    if (bracket_found) {
         deferred_logger.debug(
                 " brute force solve found low " + std::to_string(low) + " with eq_low " + std::to_string(eq_low) +
                 " high " + std::to_string(high) + " with eq_high " + std::to_string(eq_high));
     }
-    return finding_bracket;
+    return bracket_found;
+}
+
+bool WellBhpThpCalculator::
+isStableSolution(const WellState& well_state,
+                 const Well& well,
+                 const std::vector<double>& rates,
+                 const SummaryState& summaryState) const
+{
+    assert(int(rates.size()) == 3); // the vfp related only supports three phases now.
+    assert(well_.isProducer()); // only valid for producers 
+
+    static constexpr int Gas = BlackoilPhases::Vapour;
+    static constexpr int Oil = BlackoilPhases::Liquid;
+    static constexpr int Water = BlackoilPhases::Aqua;
+
+    const double aqua = rates[Water];
+    const double liquid = rates[Oil];
+    const double vapour = rates[Gas];
+    const double thp = well_.getTHPConstraint(summaryState);
+
+    const auto& controls = well.productionControls(summaryState);
+    const auto& wfr =  well_.vfpProperties()->getExplicitWFR(controls.vfp_table_number, well_.indexOfWell());
+    const auto& gfr = well_.vfpProperties()->getExplicitGFR(controls.vfp_table_number, well_.indexOfWell());
+
+    const auto& table = well_.vfpProperties()->getProd()->getTable(controls.vfp_table_number);
+    const bool use_vfpexplicit = well_.useVfpExplicit();
+
+    detail::VFPEvaluation bhp = detail::bhp(table, aqua, liquid, vapour, thp, well_.getALQ(well_state), wfr, gfr, use_vfpexplicit);
+
+    if (bhp.dflo >= 0) {
+        return true;
+    } else {    // maybe check if ipr is available
+        const auto ipr = getFloIPR(well_state, well, summaryState);
+        return bhp.dflo + 1/ipr.second >= 0; 
+    }                  
+}
+
+std::optional<double> WellBhpThpCalculator::
+estimateStableBhp(const WellState& well_state,
+                  const Well& well,
+                  const std::vector<double>& rates,
+                  const double rho,
+                  const SummaryState& summaryState) const
+{   
+    // Given a *converged* well_state with ipr, estimate bhp of the stable solution 
+    const auto& controls = well.productionControls(summaryState);
+    const double thp = well_.getTHPConstraint(summaryState);
+    const auto& table = well_.vfpProperties()->getProd()->getTable(controls.vfp_table_number);
+
+    const double aqua = rates[BlackoilPhases::Aqua];
+    const double liquid = rates[BlackoilPhases::Liquid];
+    const double vapour = rates[BlackoilPhases::Vapour];
+    double flo = detail::getFlo(table, aqua, liquid, vapour);
+    double wfr, gfr;
+    if (well_.useVfpExplicit() || -flo < table.getFloAxis().front()) {
+        wfr =  well_.vfpProperties()->getExplicitWFR(controls.vfp_table_number, well_.indexOfWell());
+        gfr = well_.vfpProperties()->getExplicitGFR(controls.vfp_table_number, well_.indexOfWell());
+    } else {
+        wfr = detail::getWFR(table, aqua, liquid, vapour);  
+        gfr = detail::getGFR(table, aqua, liquid, vapour);   
+    }
+
+    auto ipr = getFloIPR(well_state, well, summaryState);
+    
+    const double vfp_ref_depth = well_.vfpProperties()->getProd()->getTable(controls.vfp_table_number).getDatumDepth();
+
+    const double dp_hydro = wellhelpers::computeHydrostaticCorrection(well_.refDepth(), vfp_ref_depth, 
+                                                                      rho, well_.gravity());
+    auto bhp_adjusted = [this, &thp, &dp_hydro](const double bhp) {
+           return bhp - dp_hydro + getVfpBhpAdjustment(bhp, thp);
+       };
+    const auto retval = detail::intersectWithIPR(table, thp, wfr, gfr, well_.getALQ(well_state), ipr.first, ipr.second, bhp_adjusted);
+    if (retval.has_value()) {
+        // returned pair is (flo, bhp)
+        return retval.value().second;
+    } else {
+        return std::nullopt;
+    }
+}    
+
+std::pair<double, double> WellBhpThpCalculator::
+getFloIPR(const WellState& well_state,
+          const Well& well, 
+          const SummaryState& summary_state) const 
+{
+    // Convert ipr_a's and ipr_b's to our particular choice of FLO 
+    const auto& controls = well.productionControls(summary_state);
+    const auto& table = well_.vfpProperties()->getProd()->getTable(controls.vfp_table_number);
+    const auto& pu = well_.phaseUsage();
+    const auto& ipr_a = well_state.well(well_.indexOfWell()).implicit_ipr_a;
+    const double& aqua_a = pu.phase_used[BlackoilPhases::Aqua]? ipr_a[pu.phase_pos[BlackoilPhases::Aqua]] : 0.0;
+    const double& liquid_a = pu.phase_used[BlackoilPhases::Liquid]? ipr_a[pu.phase_pos[BlackoilPhases::Liquid]] : 0.0;
+    const double& vapour_a = pu.phase_used[BlackoilPhases::Vapour]? ipr_a[pu.phase_pos[BlackoilPhases::Vapour]] : 0.0;
+    const auto& ipr_b = well_state.well(well_.indexOfWell()).implicit_ipr_b;
+    const double& aqua_b = pu.phase_used[BlackoilPhases::Aqua]? ipr_b[pu.phase_pos[BlackoilPhases::Aqua]] : 0.0;
+    const double& liquid_b = pu.phase_used[BlackoilPhases::Liquid]? ipr_b[pu.phase_pos[BlackoilPhases::Liquid]] : 0.0;
+    const double& vapour_b = pu.phase_used[BlackoilPhases::Vapour]? ipr_b[pu.phase_pos[BlackoilPhases::Vapour]] : 0.0;
+    // The getFlo helper is indended to pick one or add two of the phase rates (depending on FLO-type), 
+    // but we can equally use it to pick/add the corresponding ipr_a, ipr_b  
+    return std::make_pair(detail::getFlo(table, aqua_a, liquid_a, vapour_a), 
+                          detail::getFlo(table, aqua_b, liquid_b, vapour_b));
 }
 
 #define INSTANCE(...) \
@@ -778,6 +984,7 @@ INSTANCE(DenseAd::Evaluation<double,6,0u>)
 INSTANCE(DenseAd::Evaluation<double,7,0u>)
 INSTANCE(DenseAd::Evaluation<double,8,0u>)
 INSTANCE(DenseAd::Evaluation<double,9,0u>)
+INSTANCE(DenseAd::Evaluation<double,10,0u>)
 INSTANCE(DenseAd::Evaluation<double,-1,4u>)
 INSTANCE(DenseAd::Evaluation<double,-1,5u>)
 INSTANCE(DenseAd::Evaluation<double,-1,6u>)
@@ -785,5 +992,5 @@ INSTANCE(DenseAd::Evaluation<double,-1,7u>)
 INSTANCE(DenseAd::Evaluation<double,-1,8u>)
 INSTANCE(DenseAd::Evaluation<double,-1,9u>)
 INSTANCE(DenseAd::Evaluation<double,-1,10u>)
-
+INSTANCE(DenseAd::Evaluation<double,-1,11u>)
 } // namespace Opm
