@@ -24,6 +24,7 @@
 
 #include <opm/common/ErrorMacros.hpp>
 
+#include <opm/input/eclipse/Schedule/Well/FilterCake.hpp>
 #include <opm/input/eclipse/Schedule/Well/WellBrineProperties.hpp>
 #include <opm/input/eclipse/Schedule/Well/WellConnections.hpp>
 #include <opm/input/eclipse/Schedule/Well/WellFoamProperties.hpp>
@@ -173,6 +174,11 @@ const Well& WellInterfaceGeneric::wellEcl() const
     return well_ecl_;
 }
 
+Well& WellInterfaceGeneric::wellEcl()
+{
+    return well_ecl_;
+}
+
 const PhaseUsage& WellInterfaceGeneric::phaseUsage() const
 {
     assert(phase_usage_ != nullptr);
@@ -190,8 +196,75 @@ double WellInterfaceGeneric::rsRvInj() const
     return well_ecl_.getInjectionProperties().rsRvInj;
 }
 
+void WellInterfaceGeneric::initInjMult(const std::vector<double>& max_inj_mult)
+{
+    // prev_inj_multiplier_ will stay unchanged during the time step
+    // while inj_multiplier_ might be updated during the time step
+    this->prev_inj_multiplier_ = max_inj_mult;
+    // initializing the inj_multipler_ to be 1.0
+    this->inj_multiplier_ = std::vector<double>(max_inj_mult.size(), 1.);
+}
+
+void WellInterfaceGeneric::updateInjMult(std::vector<double>& inj_multipliers, DeferredLogger& deferred_logger) const
+{
+    if (inj_multipliers.size() != this->inj_multiplier_.size()) {
+        OPM_DEFLOG_THROW(std::runtime_error,
+                         fmt::format("We do not support changing connection numbers during simulation with WINJMULT "
+                                     "for well {}", name()),
+                         deferred_logger);
+    }
+
+    inj_multipliers = this->inj_multiplier_;
+}
+
+
+
+double WellInterfaceGeneric::getInjMult(const int perf,
+                                        const double bhp,
+                                        const double perf_pres) const
+{
+    assert(!this->isProducer());
+
+    double multiplier = 1.;
+
+    const auto perf_ecl_index = this->perforationData()[perf].ecl_index;
+    const bool is_wrev = this->well_ecl_.getInjMultMode() == Well::InjMultMode::WREV;
+
+    const bool active_injmult = (is_wrev && this->well_ecl_.aciveWellInjMult()) ||
+                              this->well_ecl_.getConnections()[perf_ecl_index].activeInjMult();
+
+    if (active_injmult) {
+        const auto& injmult= is_wrev ? this->well_ecl_.getWellInjMult()
+                                                  : this->well_ecl_.getConnections()[perf_ecl_index].injmult();
+        const double pres = is_wrev ? bhp : perf_pres;
+
+        const auto frac_press = injmult.fracture_pressure;
+        const auto gradient = injmult.multiplier_gradient;
+        if (pres > frac_press) {
+            multiplier = 1. + (pres - frac_press) * gradient;
+        }
+    }
+
+    // for CIRR mode, if there is no active WINJMULT setup, we will use the previous injection multiplier,
+    // to mimic keeping the existing fracturing open
+    if (this->well_ecl_.getInjMultMode() == Well::InjMultMode::CIRR) {
+        multiplier = std::max(multiplier, this->prev_inj_multiplier_[perf_ecl_index]);
+    }
+
+    this->inj_multiplier_[perf_ecl_index] = multiplier;
+    return multiplier;
+}
+
+
+
+
 bool WellInterfaceGeneric::wellHasTHPConstraints(const SummaryState& summaryState) const
 {
+    // only wells under prediction mode can have THP constraint
+    if (!this->wellEcl().predictionMode()) {
+        return false;
+    }
+
     if (dynamic_thp_limit_) {
         return true;
     }
@@ -276,6 +349,30 @@ void WellInterfaceGeneric::setVFPProperties(const VFPProperties* vfp_properties_
     vfp_properties_ = vfp_properties_arg;
 }
 
+void WellInterfaceGeneric::setPrevSurfaceRates(WellState& well_state,
+                                               const WellState& prev_well_state) const
+    {
+        auto& ws = well_state.well(this->index_of_well_);
+        auto& ws_prev = prev_well_state.well(this->index_of_well_);
+        // The logic here is a bit fragile:
+        // We need non-zero prev_surface_rates for the purpose of providing explicit fractions
+        // (if needed) for vfp interpolation.
+        // We assume that current surface rates either are initialized from previous step
+        // or (if newly opened) from updateWellStateRates. This is fine unless well was
+        // stopped in previous step in which case it's rates will be zero. In this case,
+        // we select the previous rates of the previous well state (and hope for the best).
+        const bool zero_rates = std::all_of(ws.surface_rates.begin(), ws.surface_rates.end(),
+                [](double rate) {
+                    return rate == 0.; // TODO: should we use a threshhold for comparison?
+                } );
+
+        if (zero_rates) {
+            ws.prev_surface_rates = ws_prev.prev_surface_rates;
+        } else {
+            ws.prev_surface_rates = ws.surface_rates;
+        }
+    }
+
 void WellInterfaceGeneric::setGuideRate(const GuideRate* guide_rate_arg)
 {
     guide_rate_ = guide_rate_arg;
@@ -400,7 +497,7 @@ bool WellInterfaceGeneric::isOperableAndSolvable() const
 bool WellInterfaceGeneric::useVfpExplicit() const
 {
     const auto& wvfpexp = well_ecl_.getWVFPEXP();
-    return ((wvfpexp.explicit_lookup() && !changedToOpenThisStep())|| operability_status_.use_vfpexplicit);
+    return (wvfpexp.explicit_lookup() || operability_status_.use_vfpexplicit);
 }
 
 bool WellInterfaceGeneric::thpLimitViolatedButNotSwitched() const
@@ -473,6 +570,11 @@ bool WellInterfaceGeneric::stopppedOrZeroRateTarget(const SummaryState& summary_
 
 }
 
+void WellInterfaceGeneric::resetWellOperability()
+{
+    this->operability_status_.resetOperability();
+}
+
 double WellInterfaceGeneric::wmicrobes_() const
 {
     auto injectorType = this->well_ecl_.injectorType();
@@ -570,4 +672,124 @@ int WellInterfaceGeneric::polymerInjTable_() const
     return this->well_ecl_.getPolymerProperties().m_plymwinjtable;
 }
 
+std::pair<bool,bool> WellInterfaceGeneric::
+computeWellPotentials(std::vector<double>& well_potentials,
+                      const WellState& well_state)
+{
+    const int np = this->number_of_phases_;
+    well_potentials.resize(np, 0.0);
+
+    // Stopped wells have zero potential.
+    if (this->wellIsStopped()) {
+        return {false, false};
+    }
+    this->operability_status_.has_negative_potentials = false;
+
+    // If the well is pressure controlled the potential equals the rate.
+    bool thp_controlled_well = false;
+    bool bhp_controlled_well = false;
+    bool compute_potential = true;
+    const auto& ws = well_state.well(this->index_of_well_);
+    if (this->isInjector()) {
+        const Well::InjectorCMode& current = ws.injection_cmode;
+        if (current == Well::InjectorCMode::THP) {
+            thp_controlled_well = true;
+        }
+        if (current == Well::InjectorCMode::BHP) {
+            bhp_controlled_well = true;
+        }
+    } else {
+        const Well::ProducerCMode& current = ws.production_cmode;
+        if (current == Well::ProducerCMode::THP) {
+            thp_controlled_well = true;
+        }
+        if (current == Well::ProducerCMode::BHP) {
+            bhp_controlled_well = true;
+        }
+    }
+
+    if (!this->changed_to_open_this_step_ &&
+        (thp_controlled_well || bhp_controlled_well)) {
+        double total_rate = 0.0;
+        const double sign = this->isInjector() ? 1.0 : -1.0;
+        for (int phase = 0; phase < np; ++phase){
+            total_rate += sign * ws.surface_rates[phase];
+        }
+        // for pressure controlled wells the well rates are the potentials
+        // if the rates are trivial we are most probably looking at the newly
+        // opened well, and we therefore make the effort of computing the potentials anyway.
+        if (total_rate > 0) {
+            for (int phase = 0; phase < np; ++phase){
+                well_potentials[phase] = sign * ws.surface_rates[phase];
+            }
+            compute_potential = false;
+        }
+    }
+
+    return {compute_potential, bhp_controlled_well};
+}
+
+void WellInterfaceGeneric::
+checkNegativeWellPotentials(std::vector<double>& well_potentials,
+                            const bool checkOperability,
+                            DeferredLogger& deferred_logger)
+{
+    const double sign = this->isInjector() ? 1.0 : -1.0;
+    double total_potential = 0.0;
+    for (int phase = 0; phase < this->number_of_phases_; ++phase) {
+        well_potentials[phase] *= sign;
+        total_potential += well_potentials[phase];
+    }
+    if (total_potential < 0.0 && checkOperability) {
+        // wells with negative potentials are not operable
+        this->operability_status_.has_negative_potentials = true;
+        const std::string msg = std::string("well ") + this->name() +
+                                ": has negative potentials and is not operable";
+        deferred_logger.warning("NEGATIVE_POTENTIALS_INOPERABLE", msg);
+    }
+}
+
+void WellInterfaceGeneric::
+prepareForPotentialCalculations(const SummaryState& summary_state,
+                                WellState& well_state, 
+                                Well::InjectionControls& inj_controls,
+                                Well::ProductionControls& prod_controls) const
+{
+    const bool has_thp = this->wellHasTHPConstraints(summary_state);
+    auto& ws = well_state.well(this->index_of_well_);
+    // Modify control (only pressure constraints) and set new target if needed.
+    // Also set value for current target in state
+    if (this->isInjector()) {
+        inj_controls.clearControls();
+        inj_controls.addControl(Well::InjectorCMode::BHP);
+        if (has_thp){
+            inj_controls.addControl(Well::InjectorCMode::THP);
+        }
+        if (!(ws.injection_cmode == Well::InjectorCMode::BHP)){
+            if (has_thp){
+                ws.injection_cmode = Well::InjectorCMode::THP;
+                ws.thp = this->getTHPConstraint(summary_state);
+            } else {
+                ws.injection_cmode = Well::InjectorCMode::BHP;
+                ws.bhp = inj_controls.bhp_limit;
+            }
+        } 
+    } else { // producer
+        prod_controls.clearControls();
+        prod_controls.addControl(Well::ProducerCMode::BHP);
+        if (has_thp){
+            prod_controls.addControl(Well::ProducerCMode::THP);
+        }
+        if (!(ws.production_cmode == Well::ProducerCMode::BHP)){
+            if (has_thp){
+                ws.production_cmode = Well::ProducerCMode::THP;
+                ws.thp = this->getTHPConstraint(summary_state);
+            } else {
+                ws.production_cmode = Well::ProducerCMode::BHP;
+                ws.bhp = prod_controls.bhp_limit;
+            }
+        } 
+    }    
+}
+
 } // namespace Opm
