@@ -21,10 +21,18 @@
 #ifndef OPM_WELLSTATEFULLYIMPLICITBLACKOIL_HEADER_INCLUDED
 #define OPM_WELLSTATEFULLYIMPLICITBLACKOIL_HEADER_INCLUDED
 
-#include <opm/common/ErrorMacros.hpp>
+#include <dune/common/version.hh>
+#include <dune/common/parallel/mpihelper.hh>
 
 #include <opm/core/props/BlackoilPhases.hpp>
 
+#include <opm/common/ErrorMacros.hpp>
+
+#include <opm/input/eclipse/Schedule/Events.hpp>
+
+#include <opm/output/data/Wells.hpp>
+
+#include <opm/simulators/utils/ParallelCommunication.hpp>
 #include <opm/simulators/wells/ALQState.hpp>
 #include <opm/simulators/wells/GlobalWellInfo.hpp>
 #include <opm/simulators/wells/PerfData.hpp>
@@ -33,13 +41,6 @@
 #include <opm/simulators/wells/SingleWellState.hpp>
 #include <opm/simulators/wells/WellContainer.hpp>
 
-#include <opm/output/data/Wells.hpp>
-
-#include <opm/input/eclipse/Schedule/Events.hpp>
-
-#include <dune/common/version.hh>
-#include <dune/common/parallel/mpihelper.hh>
-
 #include <functional>
 #include <map>
 #include <optional>
@@ -108,7 +109,7 @@ public:
                 const std::vector<std::reference_wrapper<ParallelWellInfo>>& parallel_well_info,
                 const Schedule& schedule,
                 const bool handle_ms_well,
-                const size_t numCells,
+                const std::size_t numCells,
                 const std::vector<std::vector<PerforationData>>& well_perf_data,
                 const SummaryState& summary_state);
 
@@ -129,10 +130,9 @@ public:
         this->well_rates.clear();
     }
 
-    template<class Communication>
     void gatherVectorsOnRoot(const std::vector< data::Connection >& from_connections,
                              std::vector< data::Connection >& to_connections,
-                             const Communication& comm) const;
+                             const Parallel::Communication& comm) const;
 
     data::Wells
     report(const int* globalCellIdxMap,
@@ -150,11 +150,9 @@ public:
                                       const std::vector<double>& perforation_rates, const int np, const int segment, std::vector<double>& segment_rates);
 
 
-    template<class Comm>
-    void communicateGroupRates(const Comm& comm);
+    void communicateGroupRates(const Parallel::Communication& comm);
 
-    template<class Comm>
-    void updateGlobalIsGrup(const Comm& comm);
+    void updateGlobalIsGrup(const Parallel::Communication& comm);
 
     bool isInjectionGrup(const std::string& name) const {
         return this->global_well_info.value().in_injecting_group(name);
@@ -206,6 +204,12 @@ public:
         this->alq_state.reset_count();
     }
 
+    // If the ALQ has changed since the previous time step,
+    // reset current_alq and update default_alq. ALQ is used for
+    // constant lift gas injection and for gas lift optimization
+    // (THP controlled wells).
+    void updateWellsDefaultALQ(const std::vector<Well>& wells_ecl, const SummaryState& summary_state);
+
     int wellNameToGlobalIdx(const std::string &name) {
         return this->global_well_info.value().well_index(name);
     }
@@ -331,12 +335,6 @@ private:
                          const int         seg_ix,
                          const int         seg_no) const;
 
-    // If the ALQ has changed since the previous report step,
-    // reset current_alq and update default_alq. ALQ is used for
-    // constant lift gas injection and for gas lift optimization
-    // (THP controlled wells).
-
-    void updateWellsDefaultALQ(const std::vector<Well>& wells_ecl);
 
     /// Allocate and initialize if wells is non-null.
     /// Also tries to give useful initial values to the bhp() and
