@@ -19,9 +19,7 @@
 
 #include <config.h>
 
-#include <ebos/ebos.hh>
-#include <ebos/eclgenericvanguard.hh>
-#include <ebos/femcpgridcompat.hh>
+#include "TestTypeTag.hpp"
 
 #include <opm/common/utility/Serializer.hpp>
 
@@ -32,11 +30,16 @@
 
 #include <opm/output/eclipse/Inplace.hpp>
 
+#include <opm/input/eclipse/EclipseState/WagHysteresisConfig.hpp>
+
 #include <opm/material/fluidmatrixinteractions/EclHysteresisTwoPhaseLawParams.hpp>
 
 #include <opm/models/blackoil/blackoilprimaryvariables.hh>
 
-#include <opm/simulators/timestepping/AdaptiveTimeSteppingEbos.hpp>
+#include <opm/simulators/flow/FemCpGridCompat.hpp>
+#include <opm/simulators/flow/FlowGenericVanguard.hpp>
+#include <opm/simulators/flow/FlowProblem.hpp>
+#include <opm/simulators/timestepping/AdaptiveTimeStepping.hpp>
 #include <opm/simulators/timestepping/SimulatorReport.hpp>
 #include <opm/simulators/timestepping/SimulatorTimer.hpp>
 #include <opm/simulators/timestepping/TimeStepControl.hpp>
@@ -56,6 +59,19 @@
 
 #include <algorithm>
 
+namespace Opm::Properties {
+    namespace TTag {
+    struct TestRestartTypeTag {
+            using InheritsFrom = std::tuple<TestTypeTag, FlowTimeSteppingParameters>;
+        };
+    }
+
+    template<>
+    struct LinearSolverBackend<TTag::TestRestartTypeTag, TTag::FlowIstlSolverParams> {
+        using type = ISTLSolver<TTag::TestRestartTypeTag>;
+    };
+}
+
 template<class T>
 std::tuple<T,int,int> PackUnpack(T& in)
 {
@@ -98,13 +114,13 @@ TEST_FOR_TYPE(SimulatorReport)
 TEST_FOR_TYPE(SimulatorReportSingle)
 TEST_FOR_TYPE(SimulatorTimer)
 
-namespace Opm { using ATE = AdaptiveTimeSteppingEbos<Properties::TTag::EbosTypeTag>; }
-TEST_FOR_TYPE_NAMED_OBJ(ATE, AdaptiveTimeSteppingEbosHardcoded, serializationTestObjectHardcoded)
-TEST_FOR_TYPE_NAMED_OBJ(ATE, AdaptiveTimeSteppingEbosPID, serializationTestObjectPID)
-TEST_FOR_TYPE_NAMED_OBJ(ATE, AdaptiveTimeSteppingEbosPIDIt, serializationTestObjectPIDIt)
-TEST_FOR_TYPE_NAMED_OBJ(ATE, AdaptiveTimeSteppingEbosSimple, serializationTestObjectSimple)
+namespace Opm { using ATS = AdaptiveTimeStepping<Properties::TTag::TestTypeTag>; }
+TEST_FOR_TYPE_NAMED_OBJ(ATS, AdaptiveTimeSteppingHardcoded, serializationTestObjectHardcoded)
+TEST_FOR_TYPE_NAMED_OBJ(ATS, AdaptiveTimeSteppingPID, serializationTestObjectPID)
+TEST_FOR_TYPE_NAMED_OBJ(ATS, AdaptiveTimeSteppingPIDIt, serializationTestObjectPIDIt)
+TEST_FOR_TYPE_NAMED_OBJ(ATS, AdaptiveTimeSteppingSimple, serializationTestObjectSimple)
 
-namespace Opm { using BPV = BlackOilPrimaryVariables<Properties::TTag::EbosTypeTag>; }
+namespace Opm { using BPV = BlackOilPrimaryVariables<Properties::TTag::TestTypeTag>; }
 TEST_FOR_TYPE_NAMED(BPV, BlackoilPrimaryVariables)
 
 namespace Opm {
@@ -122,7 +138,7 @@ namespace Opm {
 TEST_FOR_TYPE_NAMED(HystParam, EclHysteresisTwoPhaseLawParams)
 
 namespace Opm {
-    using Disc = Opm::FvBaseDiscretization<Opm::Properties::TTag::EbosTypeTag>;
+    using Disc = Opm::FvBaseDiscretization<Opm::Properties::TTag::TestTypeTag>;
     using BVec = typename Disc::BlockVectorWrapper;
 }
 TEST_FOR_TYPE_NAMED(BVec, BlockVectorWrapper)
@@ -189,24 +205,24 @@ BOOST_AUTO_TEST_CASE(WGState)
 
 BOOST_AUTO_TEST_CASE(EclGenericVanguard)
 {
-    auto in_params = Opm::EclGenericVanguard::serializationTestParams();
-    Opm::EclGenericVanguard val1(std::move(in_params));
+    auto in_params = Opm::FlowGenericVanguard::serializationTestParams();
+    Opm::FlowGenericVanguard val1(std::move(in_params));
     Opm::Serialization::MemPacker packer;
     Opm::Serializer ser(packer);
     ser.pack(val1);
     const size_t pos1 = ser.position();
-    Opm::EclGenericVanguard::SimulationModelParams out_params;
+    Opm::FlowGenericVanguard::SimulationModelParams out_params;
     out_params.setupTime_ = 0.0;
     out_params.actionState_ = std::make_unique<Opm::Action::State>();
     out_params.udqState_ = std::make_unique<Opm::UDQState>();
     out_params.eclSchedule_ = std::make_shared<Opm::Schedule>();
     out_params.summaryState_ = std::make_unique<Opm::SummaryState>();
-    Opm::EclGenericVanguard val2(std::move(out_params));
+    Opm::FlowGenericVanguard val2(std::move(out_params));
     ser.unpack(val2);
     const size_t pos2 = ser.position();
 
-    BOOST_CHECK_MESSAGE(pos1 == pos2, "Packed size differ from unpack size for EclGenericVanguard");
-    BOOST_CHECK_MESSAGE(val1 == val2, "Deserialized EclGenericVanguard differ");
+    BOOST_CHECK_MESSAGE(pos1 == pos2, "Packed size differ from unpack size for FlowGenericVanguard");
+    BOOST_CHECK_MESSAGE(val1 == val2, "Deserialized FlowGenericVanguard differ");
 }
 
 BOOST_AUTO_TEST_CASE(EclGenericProblem)
@@ -224,7 +240,7 @@ BOOST_AUTO_TEST_CASE(EclGenericProblem)
     auto gridView = grid.leafGridView();
 #endif // HAVE_DUNE_FEM
     auto data_out
-        = Opm::EclGenericProblem<GridView, Opm::BlackOilFluidSystem<double, Opm::BlackOilDefaultIndexTraits>, double>::
+        = Opm::FlowGenericProblem<GridView, Opm::BlackOilFluidSystem<double, Opm::BlackOilDefaultIndexTraits>>::
             serializationTestObject(eclState, schedule, gridView);
     Opm::Serialization::MemPacker packer;
     Opm::Serializer ser(packer);
@@ -273,6 +289,7 @@ public:
         last_glift_opt_time_ = 5.0;
         switched_prod_groups_ = {{"test4", "test5"}};
         switched_inj_groups_ = {{{"test4", Phase::SOLVENT}, "test5"}};
+        closed_offending_wells_ = {{"test4", {"test5", "test6"}}};
     }
 
     void calcRates(const int, const int, const std::vector<double>&, std::vector<double>&) override
@@ -336,10 +353,11 @@ BOOST_AUTO_TEST_CASE(BlackoilWellModelGeneric)
 }
 
 template<class Grid, class GridView, class DofMapper, class Stencil, class Scalar>
-class EclGenericTracerModelTest : public Opm::EclGenericTracerModel<Grid,GridView,DofMapper,Stencil,Scalar> {
-    using Base = Opm::EclGenericTracerModel<Grid,GridView,DofMapper,Stencil,Scalar>;
+class GenericTracerModelTest : public Opm::GenericTracerModel<Grid,GridView,DofMapper,Stencil,Scalar>
+{
+    using Base = Opm::GenericTracerModel<Grid,GridView,DofMapper,Stencil,Scalar>;
 public:
-    EclGenericTracerModelTest(const GridView& gridView,
+    GenericTracerModelTest(const GridView& gridView,
                               const Opm::EclipseState& eclState,
                               const Dune::CartesianIndexMapper<Grid>& cartMapper,
                               const DofMapper& dofMapper,
@@ -347,21 +365,21 @@ public:
         Base(gridView, eclState, cartMapper, dofMapper, centroids)
     {}
 
-    static EclGenericTracerModelTest
+    static GenericTracerModelTest
     serializationTestObject(const GridView& gridView,
                             const Opm::EclipseState& eclState,
                             const Dune::CartesianIndexMapper<Grid>& cartMapper,
                             const DofMapper& dofMapper,
                             const std::function<std::array<double,Grid::dimensionworld>(int)> centroids)
     {
-        EclGenericTracerModelTest result(gridView, eclState, cartMapper, dofMapper, centroids);
+        GenericTracerModelTest result(gridView, eclState, cartMapper, dofMapper, centroids);
         result.tracerConcentration_ = {{1.0}, {2.0}, {3.0}};
         result.wellTracerRate_.insert({{"foo", "bar"}, 4.0});
 
         return result;
     }
 
-    bool operator==(const EclGenericTracerModelTest& rhs) const
+    bool operator==(const GenericTracerModelTest& rhs) const
     {
         if (this->tracerConcentration_.size() != rhs.tracerConcentration_.size()) {
             return false;
@@ -394,11 +412,11 @@ BOOST_AUTO_TEST_CASE(EclGenericTracerModel)
     auto gridView = grid.leafGridView();
 #endif // HAVE_DUNE_FEM
     Dune::MultipleCodimMultipleGeomTypeMapper<GridView> dofMapper(gridView, Dune::mcmgElementLayout());
-    auto data_out = EclGenericTracerModelTest<Dune::CpGrid,
-                                              GridView,
-                                              Dune::MultipleCodimMultipleGeomTypeMapper<GridView>,
-                                              Opm::EcfvStencil<double, GridView, false, false>,
-                                              double>
+    auto data_out = GenericTracerModelTest<Dune::CpGrid,
+                                           GridView,
+                                           Dune::MultipleCodimMultipleGeomTypeMapper<GridView>,
+                                           Opm::EcfvStencil<double, GridView, false, false>,
+                                           double>
         ::serializationTestObject(gridView, eclState, mapper, dofMapper, centroids);
     Opm::Serialization::MemPacker packer;
     Opm::Serializer ser(packer);
@@ -413,7 +431,7 @@ BOOST_AUTO_TEST_CASE(EclGenericTracerModel)
 
 namespace Opm {
 
-class TBatchExport : public EclTracerModel<Properties::TTag::EbosTypeTag> {
+class TBatchExport : public TracerModel<Properties::TTag::TestTypeTag> {
 public:
     using TBatch = TracerBatch<double>;
 };
@@ -426,13 +444,17 @@ namespace {
 
 struct AquiferFixture {
     AquiferFixture() {
-        using TT = Opm::Properties::TTag::EbosTypeTag;
+        using namespace Opm;
+        using TT = Properties::TTag::TestRestartTypeTag;
         const char* argv[] = {
             "test_RestartSerialization",
             "--ecl-deck-file-name=GLIFT1.DATA"
         };
-        Opm::setupParameters_<TT>(2, argv, /*registerParams=*/true);
-        Opm::EclGenericVanguard::setCommunication(std::make_unique<Opm::Parallel::Communication>());
+        AdaptiveTimeStepping<TT>::registerParameters();
+        BlackoilModelParameters<TT>::registerParameters();
+        Parameters::registerParam<TT, Properties::EnableTerminalOutput>("Do *NOT* use!");
+        setupParameters_<TT>(2, argv, /*registerParams=*/true);
+        FlowGenericVanguard::setCommunication(std::make_unique<Opm::Parallel::Communication>());
     }
 };
 
@@ -443,8 +465,8 @@ BOOST_GLOBAL_FIXTURE(AquiferFixture);
 #define TEST_FOR_AQUIFER(TYPE) \
 BOOST_AUTO_TEST_CASE(TYPE) \
 { \
-    using TT = Opm::Properties::TTag::EbosTypeTag; \
-    Opm::EclGenericVanguard::readDeck("GLIFT1.DATA"); \
+    using TT = Opm::Properties::TTag::TestRestartTypeTag; \
+    Opm::FlowGenericVanguard::readDeck("GLIFT1.DATA"); \
     using Simulator = Opm::GetPropType<TT, Opm::Properties::Simulator>; \
     Simulator sim; \
     auto data_out = Opm::TYPE<TT>::serializationTestObject(sim); \
@@ -464,8 +486,8 @@ TEST_FOR_AQUIFER(AquiferFetkovich)
 
 BOOST_AUTO_TEST_CASE(AquiferNumerical)
 {
-    using TT = Opm::Properties::TTag::EbosTypeTag;
-    Opm::EclGenericVanguard::readDeck("GLIFT1.DATA");
+    using TT = Opm::Properties::TTag::TestRestartTypeTag;
+    Opm::FlowGenericVanguard::readDeck("GLIFT1.DATA");
     using Simulator = Opm::GetPropType<TT, Opm::Properties::Simulator>;
     Simulator sim;
     auto data_out = Opm::AquiferNumerical<TT>::serializationTestObject(sim);
@@ -482,8 +504,8 @@ BOOST_AUTO_TEST_CASE(AquiferNumerical)
 
 BOOST_AUTO_TEST_CASE(AquiferConstantFlux)
 {
-    using TT = Opm::Properties::TTag::EbosTypeTag;
-    Opm::EclGenericVanguard::readDeck("GLIFT1.DATA");
+    using TT = Opm::Properties::TTag::TestRestartTypeTag;
+    Opm::FlowGenericVanguard::readDeck("GLIFT1.DATA");
     using Simulator = Opm::GetPropType<TT, Opm::Properties::Simulator>;
     Simulator sim;
     auto data_out = Opm::AquiferConstantFlux<TT>::serializationTestObject(sim);
